import copy

from chemotion_api.labImotion.items.validation.registry import SchemaRegistry
from chemotion_api.labImotion.items.validation.schemas.abstract_element_schema import generic_element_schema
from chemotion_api.labImotion.items.validation.schemas.schema_properties import properties_schema as properties

generic_schema = copy.deepcopy(generic_element_schema)
generic_schema['$id'] = "chemotion://generic/linked_element/draft-01"
generic_schema['title'] = "Schema for generic linked element"
generic_schema['properties']['properties_template']['$ref'] = 'chemotion://generic/linked_properties/draft-01'
generic_schema['properties']['properties_release']['$ref'] = 'chemotion://generic/linked_properties/draft-01'
generic_schema['required'] += ['name', 'icon_name', 'is_generic', 'klass_prefix']

generic_schema['properties'] |= {
    "name": {
        "type": "string"
    },
    "icon_name": {
        "type": "string"
    },
    "klass_prefix": {
        "type": "string"
    },
    "is_generic": {
        "type": "boolean"
    },
    "version": {
        "type": ["string", 'null']
    },
    "created_by": {
        "type": ["number", 'null']
    },
    "updated_by": {
        "type": ["number", 'null']
    }
}

generic_props_schema = copy.deepcopy(properties)
generic_props_schema['$id'] = 'chemotion://generic/linked_properties/draft-01'
generic_props_schema['title'] = "Schema for generic linked properties"

generic_props_schema['properties'] |= {
    "version": {
        "type": ["string", 'null']
    }
}

generic_props_schema['required'] = [x for x in generic_props_schema['required'] if x not in ['pkg', 'version', 'layers', 'identifier']]

SchemaRegistry.instance().register(generic_props_schema)
SchemaRegistry.instance().register(generic_schema)



from __future__ import annotations
import uuid

from chemotion_api.labImotion.items.options import ConditionOperator, FieldType
from chemotion_api.labImotion.structure import StructureRegistry

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from chemotion_api.labImotion.items.fields import SelectField, TextField, CheckboxField


class ConditionHandler:
    def __init__(self, sr: StructureRegistry):
        if not hasattr(self, '_sr'):
            self._sr = sr
        self._cond_fields = None
        self._cond_operator = None

    def add_condition(self, field: SelectField | TextField | CheckboxField, value: str | bool):
        if field.field_type == FieldType.SELECT:
            options = self._sr.get_select_options()[field.option_layers]
            if value not in options:
                def find_label():
                    for key, label in options:
                        if value == label:
                            return key
                    return None

                value = find_label()
            if value is None:
                raise ValueError(f"Value must be in {', '.join(options.values())}")


        elif field.field_type == FieldType.TEXT:
            if not isinstance(value, str):
                raise ValueError('Condition value for text fields must be a string')
        elif field.field_type == FieldType.CHECKBOX:
            if isinstance(value, bool):
                value = 'true' if value else 'false'
            if value not in ['true', 'false']:
                raise ValueError('Condition value for checkboxes must be "true" or "false"')
        else:
            raise ValueError(f"Only Fields of type [{FieldType.CHECKBOX}, {FieldType.TEXT}, {FieldType.SELECT}]")

        if self.cond_fields is None:
            self.cond_fields = []
        self.cond_fields.append({
            "id": uuid.uuid4().__str__(),
            "field": field.field,
            "label": "TEST",
            "layer": field.layer_key,
            "value": value
        })

    @property
    def cond_fields(self) -> list | None:
        return self._cond_fields

    @cond_fields.setter
    def cond_fields(self, value: list | None):
        if value is None:
            self._cond_operator = None
            self._cond_fields = None
        if not isinstance(value, list):
            raise TypeError(f"cond_fields {type(value)} not supported")
        self._cond_fields = value

    @property
    def cond_operator(self) -> ConditionOperator | None:
        if self.cond_fields is not None and self._cond_operator is None:
            return ConditionOperator.MATCH_ONE
        return self._cond_operator

    @cond_operator.setter
    def cond_operator(self, value: ConditionOperator | int | None):
        if value is None:
            if self.cond_fields is None:
                self._cond_operator = None
                return
            else:
                self._cond_operator = ConditionOperator.MATCH_ONE
        if isinstance(value, int):
            value = ConditionOperator(value)
        if not isinstance(value, ConditionOperator):
            raise TypeError(f"cond_operator {type(value)} not supported")
        self._cond_operator = value



from enum import Enum
from math import isnan
from typing import Callable, Optional, Iterator, Literal

from rdflib.plugins.sparql.algebra import analyse
from requests import Response

from chemotion_api.elements.abstract_element import AbstractElement, Analyses, AnalysesList, Segments, Segment
from datetime import datetime

from chemotion_api.elements.sample import Sample, SampleVariationType, GasPhaseObj
from chemotion_api.elements.utils import QuantityUnit, GasType
from chemotion_api.elements.schemas.reaction import schema, PURIFICATION_OPTIONS, STATUS_OPTIONS
from chemotion_api.elements.utils.converter import convert_value_to_default, VOLUME_UNITS, convert_value
from chemotion_api.generic_segments import GenericSegments
from chemotion_api.labImotion.items.options import FieldType
from chemotion_api.utils import TypedList, quill_hedging, merge_dicts, PropertyDict, UnitDict
from collections.abc import MutableSequence


class MaterialList(TypedList):
    """
    A list which accepts only :class:`chemotion_api.elements.sample.Sample`'s.
    If you add a sample using th standard list-methods a splited sample will be created.
    Then the created sample will be added to the list.
    If the element has no ID, it will be saved.

    In order to avoid the splitting and the pre saving use the 'append_no_split'
    methode.
    """

    def __init__(self, reaction: 'Reaction',
                 mat_type: Literal['starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents'],
                 *args):
        super().__init__(Sample, *args)
        self._reaction = reaction
        self._mat_type = mat_type

    def _prepare_element(self, element: Sample):
        if self._mat_type == 'starting_materials':
            return_element = element.split()
        elif self._mat_type == 'reactants':
            return_element = element.split()
            return_element.properties['short_label'] = 'reactant'
        elif self._mat_type == 'products':
            return_element = element.copy()
        elif self._mat_type == 'solvents':
            return_element = element
        elif self._mat_type == 'purification_solvents':
            return_element = element
        else:
            raise ValueError("Unknown material type.")
        return_element.reaction = self._reaction
        return_element.role_in_reaction = self._mat_type
        return_element.set_ref(False, False)
        return_element.properties['position'] = len(self)

        return return_element

    def after_append(self, element: Sample):
        try:

            element.set_amount(**element.get_amount())
        except ValueError or KeyError:
            element.set_amount(0, element.allowed_quantity_units()[0])

    def append_no_split(self, element: Sample):
        """
        Add a Sample without splitting it
        :param element: Sample to be added
        """

        self._check_element(element)
        return super(TypedList, self).append(element)


class Temperature(dict):
    """
    This object contains the  temperature-time profile, the temperature unit and a user text.
    Each entry contains a time as 'hh:mm:ss' and a temperature as integer.

    :key data: {list} the temperature-time profile
    :key userText: {str}
    :key valueUnit: {str}
    """

    def __init__(self, **kwargs):
        super().__init__(data=kwargs.get('data', []),
                         userText=kwargs.get('userText', ''),
                         valueUnit=kwargs.get('valueUnit', "°C")
                         )

    def add_time_point(self, hour: int, minute: int, second: int, temperature: float):
        """
        Adds an entry to the Temperature timeline

        :param hour: since the reaction has started
        :param minute: since the reaction has started
        :param second: since the reaction has started
        :param temperature: degrees
        """
        data = self.get('data')
        if data is None:
            self['data'] = []
            data = self['data']
        data.append(
            {'time': f'{str(hour).zfill(2)}:{str(minute).zfill(2)}:{str(second).zfill(2)}', 'value': str(temperature)})


class Reaction(AbstractElement):
    """
    A chemotion Reaction object.
    It extends the :class:`chemotion_api.elements.abstract_element.AbstractElement`

    Usage::

    >>> from chemotion_api import Instance
    >>> from chemotion_api.collection import Collection
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the reaction with ID 1
    >>> rea = instance.get_reaction(1)
    >>> # Set the real amount to 3.7 g
    >>> col_solv: Collection = instance.get_root_collection().get_or_create_collection('Solv')
    >>> # Create a new solvent CDCl3
    >>> solv = col_solv.new_solvent('CDCl3')
    >>> # Add CDCl3 as solvent to the reaction
    >>> rea.solvents.append(solv)
    >>> # Add a split of sample with ID 1 as starting material
    >>> rea.starting_materials.append(instance.get_sample(1))
    >>> # Add a new time/temperature step to the temperature timeline
    >>> rea.properties['temperature'].add_time_point(2,3,0,100)
    >>> rea.save()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._variations = None

    datetime_format = '%m/%d/%Y %H:%M:%S'

    def _set_json_data(self, json_data: dict):
        super()._set_json_data(json_data)
        self._svg_file = self.json_data.get('reaction_svg_file')

    def load_image(self) -> Response:
        """
        Loads the reaction structure as svg image

        :return: Response with the svg as content
        """

        image_url = "/images/reactions/{}".format(self._svg_file)
        res = self._session.get(image_url)
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        return res

    def properties_schema(self) -> dict:
        """
        Returns the JSON.org schema of the cleaned properties.

        :return: JSON.org schema
        """

        return schema

    @property
    def properties(self) -> PropertyDict:
        """
        The properties property contains all data which can be altered
        through the chemotion api from the main tab of the reaction.


        :key starting_materials: {:class:`chemotion_api.elements.reaction.MaterialList`}
        :key reactants:  {:class:`chemotion_api.elements.reaction.MaterialList`}
        :key products: {:class:`chemotion_api.elements.reaction.MaterialList`}
        :key solvents: {:class:`chemotion_api.elements.reaction.MaterialList`}
        :key purification_solvents: {:class:`chemotion_api.elements.reaction.MaterialList`}
        :key temperature: {:class:`chemotion_api.elements.reaction.Temperature`}
        :key timestamp_start: {datetime.datetime}
        :key timestamp_stop: {datetime.datetime}
        :key name: {str}
        :key description: {str|dict} A Guill.js text (https://quilljs.com/docs/delta)
        :key observation: {str|dict} A Guill.js text (https://quilljs.com/docs/delta)
        :key purification: {list[str]} values must be in ['Flash-Chromatography', 'TLC', 'HPLC', 'Extraction', 'Distillation', 'Dialysis', 'Filtration', 'Sublimation', 'Crystallisation', 'Recrystallisation', 'Precipitation']
        :key status: {str} value must be in ['', 'Planned', 'Running', 'Done', 'Analyses Pending', 'Successful',
                                          'Not Successful']
        :key vessel_size: {dict} has a unit {str} ('l' or 'ml') and an amount {float}

        Readonly properties:

        :key short_label: {str, readonly}
        :key tlc_solvents: {str, readonly}
        :key tlc_description: {str, readonly}
        :key reaction_svg_file: {str, readonly}
        :key role: {str, readonly}
        :key rf_value: {str, readonly}
        :key rxno: {str, readonly}
        :key literatures: {str, readonly}
        :key variations: {str, readonly} Can be used with the reaction class property variations

        :return: Element properties
        """
        return super().properties

    @property
    def vessel_size_in_l(self):
        if self.properties['vessel_size']['amount'] is None:
            return 0
        return convert_value_to_default(unit=self.properties['vessel_size']['unit'],
                                        value=self.properties['vessel_size']['amount'])

    @property
    def variations(self) -> 'Variations':
        if self._variations is None:
            self._variations = Variations(self)

        return self._variations

    @property
    def starting_materials(self) -> MaterialList:
        return self.properties['starting_materials']

    @property
    def reactants(self) -> MaterialList:
        return self.properties['reactants']

    @property
    def products(self) -> MaterialList:
        return self.properties['products']

    @property
    def solvents(self) -> MaterialList:
        return self.properties['solvents']

    @property
    def purification_solvents(self) -> MaterialList:
        return self.properties['purification_solvents']

    @property
    def reference_sample(self) -> Optional[Sample]:
        first_sample = None
        for mat in [*self.starting_materials, *self.reactants]:
            if first_sample is None:
                first_sample = mat
            if mat.properties['reference']:
                return mat
        if mat is not None:
            mat.properties['reference'] = True
        return mat

    @reference_sample.setter
    def reference_sample(self, value: Sample | None):
        id = None if value is None else value.id
        for mat in [*self.starting_materials, *self.reactants]:
            mat.set_ref(mat.id == id, False)
        for mat in [*self.products, *self.solvents, *self.purification_solvents]:
            mat.set_ref(mat.id == id, False)
        self.update_equivalent()

    @property
    def gas_schema(self) -> bool:
        return self._properties.get('gaseous', False)

    @gas_schema.setter
    def gas_schema(self, gaseous: bool):
        self._properties['gaseous'] = gaseous
        if not gaseous:
            for mat in [*self.starting_materials, *self.reactants, *self.products, *self.solvents,
                        *self.purification_solvents]:
                mat.set_gas_type('off', False)
        self.update_equivalent()

    def get_feedstock(self) -> Sample | None:
        return self.get_gas_component(GasType.FEEDSTOCK)

    def get_catalyst(self) -> Sample | None:
        return self.get_gas_component(GasType.CATALYST)

    def get_gas_component(self, gt: GasType):
        if self.gas_schema:
            for mat in [*self.starting_materials, *self.reactants]:
                if mat.gas_type == gt:
                    return mat
        return None

    def update_amount(self):
        for mat in [*self.starting_materials, *self.reactants, *self.products, *self.solvents,
                    *self.purification_solvents]:
            mat.update_amount()
        for var in self.variations:
            var.update_amount()

    def update_equivalent(self):
        for mat in [*self.starting_materials, *self.reactants, *self.products, *self.solvents,
                    *self.purification_solvents]:
            mat.update_equivalent()
        for var in self.variations:
            var.update_equivalent()

    def _parse_properties(self) -> dict:
        reaction_elements = {}
        for reaction_elm_names in ['starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']:
            obj_list = self.json_data[reaction_elm_names]
            temp = []
            for sample in obj_list:
                sample_obj = Sample(self._generic_segments, self._session, sample)
                sample_obj.reaction = self
                temp.append(sample_obj)
            reaction_elements[reaction_elm_names] = MaterialList(self, reaction_elm_names, temp)

        try:
            timestamp_start = datetime.strptime(self.json_data.get('timestamp_start'), self.datetime_format)
        except:
            timestamp_start = None
        try:
            timestamp_stop = datetime.strptime(self.json_data.get('timestamp_stop'), self.datetime_format)
        except:
            timestamp_stop = None
        return reaction_elements | {
            'timestamp_start': timestamp_start,
            'timestamp_stop': timestamp_stop,
            'description': self.json_data.get('description'),
            'name': self.json_data.get('name'),
            'observation': self.json_data.get('observation'),
            'purification': self.json_data.get('purification'),
            'dangerous_products': self.json_data.get('dangerous_products'),
            'conditions': self.json_data.get('conditions'),
            'rinchi_long_key': self.json_data.get('rinchi_long_key'),
            'rinchi_web_key': self.json_data.get('rinchi_web_key'),
            'rinchi_short_key': self.json_data.get('rinchi_short_key'),
            'duration': self.json_data.get('duration'),
            'rxno': self.json_data.get('rxno'),
            'temperature': Temperature(**self.json_data.get('temperature', {})),
            'status': self.json_data.get('status'),
            'vessel_size': self.json_data.get('vessel_size'),
            'gaseous': self.json_data.get('gaseous')
            # 'tlc_solvents': self.jreaction.properties['reactants']son_data.get('tlc_solvents'),
            # 'tlc_description': self.json_data.get('tlc_description'),
            # 'rf_value': self.json_data.get('rf_value'),
        }

    def _clean_properties_data(self, serialize_data: dict | None = None) -> dict:
        if serialize_data is None:
            serialize_data = {}
        serialize_data['materials'] = {}
        for reaction_elm_names in ['starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']:
            temp_json_sample = self.json_data[reaction_elm_names]
            serialize_data['materials'][reaction_elm_names] = []
            for sample in self.properties[reaction_elm_names]:
                origen = next((x for x in temp_json_sample if x['id'] == sample.id), {})
                serialize_data['materials'][reaction_elm_names].append(origen | sample.clean_data())

        try:
            timestamp_start = self.properties.get('timestamp_start').strftime(self.datetime_format)
        except:
            timestamp_start = ''
        try:
            timestamp_stop = self.properties.get('timestamp_stop').strftime(self.datetime_format)
        except:
            timestamp_stop = ''
        serialize_data['name'] = self.properties.get('name')
        serialize_data['description'] = quill_hedging(self.properties.get('description'), 'Description')
        serialize_data['dangerous_products'] = self.properties.get('dangerous_products')
        serialize_data['conditions'] = self.properties.get('conditions')
        serialize_data['duration'] = self.properties.get('duration')
        serialize_data |= self._calc_duration()
        serialize_data['timestamp_start'] = timestamp_start
        serialize_data['timestamp_stop'] = timestamp_stop
        serialize_data['temperature'] = self.properties.get('temperature')
        serialize_data['observation'] = quill_hedging(self.properties.get('observation'), 'Observation')

        serialize_data['status'] = self.properties.get('status')
        if self.properties.get('status') in STATUS_OPTIONS:
            serialize_data['status'] = self.properties.get('status')
        else:
            serialize_data['status'] = self.json_data.get('status')

        if self.properties.get('purification') is list:
            serialize_data['purification'] = [x for x in self.properties.get('purification') if
                                              x in PURIFICATION_OPTIONS]
        else:
            serialize_data['purification'] = self.json_data.get('purification')

        if self.properties.get('vessel_size') is not None:
            serialize_data['vessel_size'] = self.properties.get('vessel_size')

        serialize_data['role'] = self.properties.get('role', '')
        serialize_data['gaseous'] = self.properties.get('gaseous')
        serialize_data['tlc_solvents'] = self.json_data.get('tlc_solvents')
        serialize_data['tlc_description'] = self.json_data.get('tlc_description')
        serialize_data['reaction_svg_file'] = self.json_data.get('reaction_svg_file')

        serialize_data['rf_value'] = self.json_data.get('rf_value')
        serialize_data['rxno'] = self.json_data.get('rxno', '')
        serialize_data['short_label'] = self.json_data.get('short_label')
        serialize_data['literatures'] = self.json_data.get('literatures')

        serialize_data['research_plans'] = self.json_data.get('research_plans')
        serialize_data['user_labels'] = self.json_data.get('user_labels')
        serialize_data['solvent'] = self.json_data.get('solvent')

        serialize_data['variations'] = self._clean_variations()

        return serialize_data

    def _calc_duration(self):
        a, b = self.properties.get('timestamp_stop'), self.properties.get('timestamp_start')
        if not isinstance(a, datetime) or not isinstance(b, datetime):
            return {
                'durationDisplay': self.json_data.get('durationDisplay'),
                'durationCalc': self.json_data.get('durationCalc')
            }
        c = a - b

        h = int(c.seconds / (60 * 60))
        m = int(c.seconds % (60 * 60) / 60)
        s = c.seconds % 60
        text = []
        total_unit = None
        total_time = 0
        total_factor = 0
        for (time, unit, factor) in ((c.days, 'day', 1), (h, 'hour', 24), (m, 'minute', 60), (s, 'second', 60)):
            total_factor *= factor
            if time > 0:
                if total_unit is None:
                    total_unit = unit + "(s)"
                    total_factor = 1
                total_time += time / total_factor
                text.append(f"{time} {unit}{'s' if time > 1 else ''}")
        return {'durationCalc': ' '.join(text),
                'durationDisplay': {
                    "dispUnit": total_unit,
                    "dispValue": f"{int(total_time)}",
                    "memUnit": total_unit,
                    "memValue": "{:0.15f}".format(total_time)
                }
                }

    def on_properties_change(self, dict_obj, name, value):
        if name[-1] == 'gaseous':
            self.gas_schema = value[-1]
            return False
        if name[-1] == 'vessel_size':
            if len(name) > 1:
                dict_obj[-1][name[-1]] = value[-1]
            elif isinstance(value[-1], int or float):
                dict_obj[-1][name[-1]]['amount'] = value[-1]
            elif VOLUME_UNITS(value[-1]):
                dict_obj[-1][name[-1]]['unit'] = value[-1]
            self.update_amount()
            return False
        return True

    def _clean_variations(self) -> list:
        cleand_data = self.variations.clean()
        self._variations = None
        return cleand_data


class SampleVariation(Sample):

    def __init__(self, reaction: Reaction, sample: Sample, variation_data: dict):
        params = sample.copy_params()
        params['avoid_load'] = True
        super().__init__(**params)
        if variation_data is None:
            self._variation_data = {}
        else:
            self._variation_data = variation_data
        self._reaction = reaction
        self._sample = sample
        self._props = None
        self._segments = None

    @property
    def gas_phase_data(self) -> GasPhaseObj:
        if not isinstance(self._props.get('gas_phase_data'), GasPhaseObj):
            gas_phase_data = self._sample._properties.get('gas_phase_data').copy_with_sample(
                self)
            gas_phase_data.set_time(**self._variation_data.get('duration', {'value': gas_phase_data.time_in_h}))
            gas_phase_data.set_temperature(
                **self._variation_data.get('temperature', {'value': gas_phase_data.temperature_in_k}))
            gas_phase_data.ppm = self._variation_data.get('concentration', {'value': gas_phase_data.ppm})[
                'value']
            self._props['gas_phase_data'] = gas_phase_data
        return self._props['gas_phase_data']

    def _extract_amount(self):
        self._props = {}
        role = self._sample.role_in_reaction
        amount_key = 'real_amount' if role == 'products' else 'target_amount'
        if role == 'products':
            self._props[amount_key] = UnitDict(**self._variation_data.get('mass', self._sample.get_amount()))
            if self._sample.gas_type == GasType.GAS:
                self._props['gas_phase_data'] = self.gas_phase_data
            self._props['equivalent'] = self._variation_data.get('yield', {}).get('value')
        else:
            if self._is_solvent:
                self._props[amount_key] = UnitDict(**self._variation_data.get('volume', self._sample.get_amount()))
            else:
                self._props[amount_key] = UnitDict(**self._variation_data.get('amount', self._sample.get_amount()))
                self._props['equivalent'] = self._variation_data.get('equivalent', {}).get('value')
        if self._props.get('equivalent') is None:
            self._props['equivalent'] = self._sample.equivalent

    @property
    def sample(self) -> Sample:
        return self._sample

    @property
    def _properties(self) -> dict:
        variation_keys = ["target_amount", "real_amount", "equivalent", "gas_phase_data"]
        if self._props is None:
            self._extract_amount()

        self._props |= {k: v for k, v in self._sample._properties.items() if k not in variation_keys}
        return self._props

    @_properties.setter
    def _properties(self, props: dict):
        pass

    def clean(self):
        self.update_equivalent()
        role = self.role_in_reaction
        amount_data = {}
        if role == 'products':
            amount_data['yield'] = {'value': self.equivalent, 'unit': None}
        elif role != 'solvents':
            amount_data['equivalent'] = {'value': self.equivalent, 'unit': None}
        if self.gas_type == GasType.GAS:
            amount_data['duration'] = {'value': convert_value(self.gas_phase_data.time_in_h, 'h', 's'),
                                       'unit': 'Second(s)'}
            amount_data['temperature'] = {'value': convert_value(self.gas_phase_data.temperature_in_k, 'K', '°C'),
                                          'unit': '°C'}
            amount_data['concentration'] = {'value': self.gas_phase_data.ppm, 'unit': 'ppm'}
            amount_data['turnoverFrequency'] = {'value': self.ton_per_h, 'unit': None}
            amount_data['turnoverNumber'] = {'value': self.ton, 'unit': None}
        if role != 'solvents':
            amount_data['mass'] = {'value': self.gram, 'unit': 'g'}
            amount_data['amount'] = {'value': self.mol, 'unit': 'mol'}
        amount_data['volume'] = {'value': self.liter, 'unit': 'l'}
        for key, value in amount_data.items():
            if amount_data[key]['value'] is None or isnan(amount_data[key]['value']):
                amount_data[key]['value'] = 0

        aux = self._variation_data.get('aux', {})
        return amount_data | {
            'aux': {
                "coefficient": self.properties['coefficient'],
                "isReference": self.properties['reference'],
                "loading": aux.get('loading'),
                "purity": self.properties['purity'],
                "density": self.properties['density'],
                "molarity": self.properties['molarity']['value'],
                "molecularWeight": self.molecule['molecular_weight'],
                "sumFormula": self.molecule['sum_formular'],
                "gasType": self.gas_type.value,
                "vesselVolume": self.vessel_size,
                "materialType": SampleVariationType.from_role_name(role).value
            }
        }

    def save(self, *args, **kwargs):
        raise NotImplementedError(f'This sample is only a Variation of {self.short_label} and cannot be saved!')



class SegmentVariation(Segment):
    allowed_fields = [FieldType.INTEGER, FieldType.TEXT, FieldType.SELECT, FieldType.SYSTEM_DEFINED]

    def __init__(self, generic_segment: dict, segment_data: Optional[dict], variation_data: list):
        super().__init__(generic_segment, segment_data)
        if segment_data is None:
            self.load()
        for values in variation_data:
            ln, fn = self._validator.convert_keys_to_labels(values['layer'], values['field'])
            self.layer[ln][fn] = values['value']

    def clean(self):
        values = {}

        for layer_label, layer_value in self.layer.items():
            for field_label, field_value in layer_value.items():
                if self._validator.field_type(layer_label, field_label) in self.allowed_fields:
                    lk, fk = self._validator.convert_labels_to_keys(layer_label, field_label)
                    if isinstance(field_value, PropertyDict or dict):
                        field_value = field_value['value']
                    values[f'{self._generic_segment['label']}___{lk}___{fk}'] = {'value': field_value}
        return values


class SegmentsVariation(Segments):

    def __init__(self, generic_segments: GenericSegments, element_type: str, segments_json: list[dict],
                 variation_json: dict):
        if segments_json is None:
            segments_json = []
        if variation_json is None:
            self._variation_json = {}
        else:
            self._variation_json = variation_json
        super().__init__(generic_segments, element_type, segments_json)

    def _populate_segments(self, segments_json):
        for seg in self._all_seg_classes:
            variation_json = []
            seg_label = seg.get('label')
            for k, x in self._variation_json.items():
                [segment, layer, field] = k.split('___')
                if segment == seg_label:
                    variation_json.append({'layer': layer, 'field': field, 'value': x})
            self.data[seg_label] = SegmentVariation(seg,
                                                    next((x for x in segments_json if
                                                          x.get('segment_klass_id') == seg.get('id')),
                                                         None),
                                                    variation_json)
    def clean(self):
        cleaned_value = {}
        for segment in self.data.values():
            if isinstance(segment, SegmentVariation):
                cleaned_value |= segment.clean()
        return cleaned_value


class Variation(Reaction):
    def __init__(self, reaction: Reaction, variation_data: dict):
        params = reaction.copy_params()
        params['avoid_load'] = True
        super().__init__(**params)
        self._reaction = reaction
        self.id = -1
        self._variation_data = variation_data
        self._materials = {}
        self._props = None
        self.populate()
        self._segments = None

    def populate(self):
        self.id = self._variation_data['id']
        for mat_type_name in ['starting_materials', 'reactants', 'products', 'solvents']:
            mat_type = SampleVariationType.from_role_name(mat_type_name)
            self._materials[mat_type] = [
                self.sample_variation_factory(x, mat_type) for x in
                getattr(self._reaction, mat_type_name)]

    @property
    def segments(self) -> Optional[Segments]:

        if self._segments is None:
            segment_variation_data = self._variation_data.get('segmentData', {})
            segment_data = self.reaction.json_data.get('segments', [])

            self._segments = SegmentsVariation(self._generic_segments,
                                               self._reaction.element_type,
                                               segment_data,
                                               segment_variation_data)
        return self._segments

    def sample_variation_factory(self, sample: Sample, sample_type: SampleVariationType) -> SampleVariation:
        json_data = None
        for key, data in self._variation_data.get(sample_type.value, {}).items():
            if key == str(sample.id):
                json_data = data

        return SampleVariation(self, sample, json_data)

    def update_and_clean(self):
        return self.clean()

    def clean(self):
        for mat_type_name in ['starting_materials', 'reactants', 'products', 'solvents']:
            mat_type = SampleVariationType.from_role_name(mat_type_name)
            self._variation_data[mat_type.value] = {}
            for mat in self._materials[mat_type]:
                self._variation_data[mat_type.value][mat.id] = mat.clean()
        self._variation_data['segmentData'] = self.segments.clean()
        return self._variation_data

    @property
    def reaction(self) -> Reaction:
        return self._reaction

    @property
    def notes(self) -> str:
        return self._variation_data['metadata']['notes']

    @notes.setter
    def notes(self, value: str):
        self._variation_data['metadata']['notes'] = value

    @property
    def temperature_in_c(self) -> float:
        return convert_value(**self._variation_data['properties']['temperature'], out_unit='°C')

    @temperature_in_c.setter
    def temperature_in_c(self, value: float):
        self._variation_data['properties']['temperature']['value'] = value

    def set_temperature(self, value, unit='°C'):
        self._variation_data['properties']['temperature']['value'] = convert_value(value, unit, '°C')

    @property
    def duration_in_s(self) -> float:
        return convert_value(**self._variation_data['properties']['duration'], out_unit='s')

    @duration_in_s.setter
    def duration_in_s(self, value: float):
        self._variation_data['properties']['duration']['value'] = value

    @property
    def variations(self):
        return []

    def set_duration(self, value, unit='s'):
        self._variation_data['properties']['duration']['value'] = convert_value(value, unit, 's')

    def link_analyses(self, analyses: Analyses):
        try:
            int(analyses.id)
            self._variation_data['metadata']['analyses'].append(analyses.id)
        except ValueError:
            raise ValueError("Analyses has no id. please save the Reaction first")

    def get_analyses(self) -> Iterator[Analyses]:
        for ana_id in self._variation_data['metadata']['analyses']:
            yield self._reaction._analyses.by_id(ana_id)

    @property
    def _properties(self) -> dict:
        variation_keys = ["starting_materials", "reactants", "products", "solvents"]
        if self._props is None:
            self._props = {k: getattr(self, k) for k in variation_keys}

        self._props |= {k: v for k, v in self._reaction._properties.items() if k not in variation_keys}
        return self._props

    @_properties.setter
    def _properties(self, props: dict):
        pass

    @property
    def starting_materials(self) -> list[SampleVariation]:
        return self._materials[SampleVariationType.STARTING_MATERIAL]

    @property
    def reactants(self) -> list[SampleVariation]:
        return self._materials[SampleVariationType.REACTANT]

    @property
    def products(self) -> list[SampleVariation]:
        return self._materials[SampleVariationType.PRODUCT]

    @property
    def solvents(self) -> list[SampleVariation]:
        return self._materials[SampleVariationType.SOLVENTS]


class Variations(MutableSequence):
    def __init__(self, reaction: 'Reaction'):
        # Initialize the list with given items
        self._data = list()
        self._reaction = reaction
        self._highest_id = 0
        self.populate(reaction.json_data['variations'])

    def populate(self, json_data: dict):
        for variation in json_data:
            v = Variation(self._reaction, variation)
            self.append(v)
            self._highest_id = max(self._highest_id, variation['id'])

    def remove(self, value: Variation) -> bool:
        try:
            idx = next(idx for idx, v in enumerate(self) if v.id == value.id)
            self.pop(idx)
        except StopIteration:
            return False
        return True

    def add_new(self) -> Variation:
        self._highest_id += 1
        variation = self.empty_variation(self._highest_id)
        v = Variation(self._reaction, variation)
        self.append(v)
        return v

    @classmethod
    def empty_variation(cls, highest_id: int) -> dict:
        return {
            "id": highest_id,
            "properties": {
                "temperature": {
                    "value": 0,
                    "unit": "°C"
                },
                "duration": {
                    "value": 0,
                    "unit": "Second(s)"
                }
            },
            "metadata": {
                "notes": "",
                "analyses": []
            },
            "reactants": {},
            "products": {},
            "solvents": {},
            "startingMaterials": {}
        }

    def sort(self):
        self._data = sorted(self._data, key=lambda v: v.id)

    def clean(self):
        return [variation.update_and_clean() for variation in self._data]

    def _prepare_variation(self, json_data: dict):
        empty_variation = self.empty_variation(self._highest_id)
        return merge_dicts(empty_variation, json_data)

    def __getitem__(self, index):
        # Retrieve the item at the given index
        return self._data[index]

    def __setitem__(self, index, value):
        # Set the item at the given index
        self._data[index] = value
        self.sort()

    def __delitem__(self, index):
        # Delete the item at the given index
        del self._data[index]

    def insert(self, index, value):
        # Insert an item at the given index
        self._data.insert(index, value)
        self.sort()

    def __len__(self):
        # Return the length of the list
        return len(self._data)

    def __repr__(self):
        # Return a string representation of the list
        return f"{repr(self._data)}"



import json
import os
import re
import string
from typing import Callable

from chemotion_api.labImotion.icons import all_icons
from chemotion_api.labImotion.items.options import ElementType
from chemotion_api.labImotion.items.validation import validate_generic_element, validate_generic_dataset, \
    validate_generic_segment, SchemaRegistry
from chemotion_api.labImotion.base_generic_obj import GenericObj
from chemotion_api.connection import Connection
from urllib.parse import urlparse, unquote


class GenericElement(GenericObj):
    url = "/api/v1/generic_elements/klasses_all.json"
    url_create = "/api/v1/generic_elements/create_element_klass"
    url_upload = '/api/v1/generic_elements/upload_klass'
    url_activate = "/api/v1/generic_elements/de_activate_klass"
    url_save = "/api/v1/generic_elements/update_element_klass"
    url_save_template = "/api/v1/generic_elements/update_template"

    element_type = ElementType.ELEMENT

    @staticmethod
    def preprocess_icon_name(icon_name: str) -> str:
        fa_prefix = 'fa fa-'
        if not isinstance(icon_name, str):
            raise TypeError(f"icon_name {type(icon_name)} not supported")
        if icon_name.startswith(fa_prefix):
            icon_name = icon_name[len(fa_prefix):]
        if icon_name not in all_icons:
            raise TypeError(f"icon_name is no valid icon name check https://fontawesome.com/v4/icons/")
        return fa_prefix + icon_name

    def __init__(self, json_data: dict, connection: Connection):
        super().__init__(json_data, connection)
        self._icon_name = json_data["icon_name"]

    def to_dict(self) -> dict:
        super().to_dict()
        self._json_data |= {
            'icon_name': self._icon_name,
            'name': self.name
        }
        return self._json_data

    @property
    def icon_name(self) -> str:
        return self._icon_name

    @icon_name.setter
    def icon_name(self, value: str):
        self._icon_name = self.preprocess_icon_name(value)

    @property
    def name(self) -> str:
        return self._json_data.get("name", '')

    @name.setter
    def name(self, value: str):
        if self.name != '':
            raise ValueError(f"name {self.name} already set")
        if not isinstance(value, str):
            raise TypeError(f"name {type(value)} not supported")
        self._json_data['name'] = value

    @classmethod
    def validator(cls) -> Callable[[dict], None]:
        return validate_generic_element


class GenericSegment(GenericObj):
    url = "/api/v1/segments/list_segment_klass.json"
    url_create = "/api/v1/segments/create_segment_klass"
    url_upload = '/api/v1/segments/upload_klass'
    url_activate = "/api/v1/generic_elements/de_activate_klass"
    url_save = "/api/v1/segments/update_segment_klass"
    url_save_template = "/api/v1/generic_elements/update_template"

    element_type = ElementType.SEGMENT

    def __init__(self, json_data: dict, connection: Connection):
        super().__init__(json_data, connection)
        self._element_klass = GenericElement(json_data["element_klass"], connection)
        pass

    @property
    def element_klass(self):
        return self._element_klass

    def to_dict(self) -> dict:
        dict_val = super().to_dict()
        dict_val |= {
            'element_klass': self._element_klass.to_dict()
        }
        return dict_val

    @classmethod
    def validator(cls) -> Callable[[dict], None]:
        return validate_generic_segment


class GenericDataset(GenericObj):
    url = "/api/v1/generic_dataset/list_dataset_klass.json"
    url_activate = "/api/v1/generic_elements/de_activate_klass"
    url_save_template = "/api/v1/generic_elements/update_template"
    element_type = ElementType.DATASET

    def __init__(self, json_data: dict, connection: Connection):
        super().__init__(json_data, connection)
        self._ols_term_id = json_data["ols_term_id"]

    @property
    def ols_term_id(self) -> str:
        return self._ols_term_id

    @ols_term_id.setter
    def ols_term_id(self, value: str):
        self._ols_term_id = value

    @classmethod
    def validator(cls) -> Callable[[dict], None]:
        return validate_generic_dataset


class GenericManager():
    def __init__(self, connection: Connection):
        self._connection = connection

    def load_all_elements(self) -> list[GenericElement]:
        return GenericElement.load_all(self._connection)

    def load_all_segments(self) -> list[GenericSegment]:
        return GenericSegment.load_all(self._connection)

    def load_all_datasets(self) -> list[GenericDataset]:
        return GenericDataset.load_all(self._connection)

    def get_element(self, **fiter_args) -> GenericElement:
        return self._get_single_generic_entry(GenericElement.load_all(self._connection), **fiter_args)

    def get_segment(self, **fiter_args) -> GenericSegment:
        return self._get_single_generic_entry(GenericSegment.load_all(self._connection), **fiter_args)

    def get_dataset(self, **fiter_args) -> GenericDataset:
        return self._get_single_generic_entry(GenericDataset.load_all(self._connection), **fiter_args)

    def _get_single_generic_entry(self, generic_element: list[GenericDataset | GenericSegment | GenericElement],
                                  **fiter_args) -> GenericDataset | GenericSegment | GenericElement:
        def filter(element):
            for key, item in fiter_args.items():
                if not hasattr(element, key) or getattr(element, key) != item:
                    return False
            return True

        filtered = [x for x in generic_element if filter(x)]
        if len(filtered) >= 2:
            raise KeyError("Multiple results!")
        if len(filtered) == 0:
            raise KeyError("No element found!")
        return filtered[0]


    def upload_elements_json(self, file_path: str) -> list[GenericElement]:
        return GenericElement.upload(self._connection, file_path)

    def upload_segment_json(self, file_path: str) -> list[GenericSegment]:
        return GenericSegment.upload(self._connection, file_path)

    def create_new_element(self, label: str, icon_name: str, description: str = '', prefix: str = '', name: str = '') -> \
            GenericElement:
        all_elements = self.load_all_elements()
        all_names = [x.name for x in all_elements]
        if name == '':
            name = t_name = re.sub(r'[^a-z]', '', label.lower())[:4]
            i = 0
            while name in all_names:
                name = t_name + string.ascii_lowercase[i]
                i += 1
        if prefix == '':
            prefix = re.sub(r'[^A-Z]', '', label.upper())[:4]
        icon_name = GenericElement.preprocess_icon_name(icon_name)
        res = GenericElement.create_new(self._connection, klass_prefix=prefix, label=label, icon_name=icon_name,
                                         name=name,
                                         desc=description)
        return self._get_single_generic_entry(res, name=name)

    def create_new_segment(self, label: str, description: str, element: GenericElement) -> GenericSegment:
        all_elements = self.load_all_elements()
        all_names = [(x.id, x.identifier) for x in all_elements if x.is_active]
        if (element.id, element.identifier) not in all_names:
            raise ValueError("Element is not active or not available")

        for seg in self.load_all_segments():
            if seg.label == label:
                raise ValueError("Segment with the same label already exists!")

        res = GenericSegment.create_new(self._connection, element_klass=element.id, label=label,
                                         desc=description)

        return self._get_single_generic_entry(res, label=label)

    def write_all_validation_schemas(self):
        cwd = os.getcwd()

        for (schema_key, schema_entry) in SchemaRegistry.instance().registry:
            parsed_url = urlparse(schema_key)
            file_path = unquote(parsed_url.path)
            if file_path.startswith('/'):
                file_path = file_path[1:]  # Remove leading slash
            file_path = os.path.join(cwd, 'schemas', file_path)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path + '.json', 'w', encoding='utf-8') as f:
                f.write(json.dumps(schema_entry.contents, indent=4))



from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

properties_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/properties/draft-01",
    "title": "Schema for generic element properties",
    "type": "object",
    "properties": {
        "pkg": {
            "type": "object",
            "properties": {
                "eln": {
                    "type": "object",
                    "properties": {
                        "version": {
                            "type": ["string", "number"]
                        },
                        "base_revision": {
                            "type": ["string", "number", "null"]
                        },
                        "current_revision": {
                            "type": ["string", "number", "null"]
                        }
                    },
                    "required": [
                        "version",
                        "base_revision",
                        "current_revision"
                    ]
                },
                "name": {
                    "type": "string"
                },
                "version": {
                    "type": "string"
                },
                "labimotion": {
                    "type": "string"
                }
            },
            "required": [
                "eln",
                "labimotion"
            ]
        },
        "uuid": {
            "type": "string"
        },
        "klass": {
            "type": "string"
        },
        "layers": {
            "type": "object",
            "patternProperties": {
                "^[a-z][a-zA-Z_]{1,20}[a-z]$": {
                    "$ref": "chemotion://generic/layer/draft-01"
                }
            },
            "additionalProperties": False
        },
        "version": {
            "type": "string"
        },
        "flowObject": {
            "type": "object",
            "properties": {
                "edges": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "id": {
                                "type": "string"
                            },
                            "label": {
                                "type": "string"
                            },
                            "source": {
                                "type": "string"
                            },
                            "target": {
                                "type": "string"
                            },
                            "animated": {
                                "type": "boolean"
                            },
                            "markerEnd": {
                                "type": "object",
                                "properties": {
                                    "type": {
                                        "type": "string"
                                    }
                                },
                                "required": [
                                    "type"
                                ]
                            },
                            "sourceHandle": {},
                            "targetHandle": {}
                        },
                        "required": [
                            "id",
                            "label",
                            "source",
                            "target",
                            "animated",
                            "markerEnd",
                            "sourceHandle",
                            "targetHandle"
                        ]
                    }
                },
                "nodes": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "id": {
                                "type": "string"
                            },
                            "data": {
                                "type": "object",
                                "properties": {
                                    "label": {
                                        "type": "string"
                                    },
                                    "lKey": {
                                        "type": "string"
                                    },
                                    "layer": {
                                        "type": "object",
                                        "properties": {
                                            "wf": {
                                                "type": "boolean"
                                            },
                                            "key": {
                                                "type": "string"
                                            },
                                            "cols": {
                                                "type": "number"
                                            },
                                            "color": {
                                                "type": "string"
                                            },
                                            "label": {
                                                "type": "string"
                                            },
                                            "style": {
                                                "type": "string"
                                            },
                                            "fields": {
                                                "type": "array",
                                                "items": {}
                                            },
                                            "position": {
                                                "type": "number"
                                            },
                                            "timeRecord": {
                                                "type": "string"
                                            },
                                            "wf_position": {
                                                "type": "number"
                                            }
                                        },
                                        "required": [
                                            "wf",
                                            "key",
                                            "cols",
                                            "color",
                                            "label",
                                            "style",
                                            "fields",
                                            "position",
                                            "timeRecord",
                                            "wf_position"
                                        ]
                                    }
                                },
                                "required": []
                            },
                            "type": {
                                "type": "string"
                            },
                            "width": {
                                "type": "number"
                            },
                            "height": {
                                "type": "number"
                            },
                            "position": {
                                "type": "object",
                                "properties": {
                                    "x": {
                                        "type": "number"
                                    },
                                    "y": {
                                        "type": "number"
                                    }
                                },
                                "required": [
                                    "x",
                                    "y"
                                ]
                            },
                            "deletable": {
                                "type": "boolean"
                            },
                            "positionAbsolute": {
                                "type": "object",
                                "properties": {
                                    "x": {
                                        "type": "number"
                                    },
                                    "y": {
                                        "type": "number"
                                    }
                                },
                                "required": [
                                    "x",
                                    "y"
                                ]
                            }
                        },
                        "required": [
                            "id",
                            "data",
                            "type",
                            "width",
                            "height",
                            "position",
                            "positionAbsolute"
                        ]
                    }
                },
                "viewport": {
                    "type": "object",
                    "properties": {
                        "x": {
                            "type": "number"
                        },
                        "y": {
                            "type": "number"
                        },
                        "zoom": {
                            "type": "number"
                        }
                    },
                    "required": [
                        "x",
                        "y",
                        "zoom"
                    ]
                }
            },
            "required": [
                "edges",
                "nodes",
                "viewport"
            ]
        },
        "identifier": {
            "type": ["string", "null"]
        },
        "select_options": {
            "$ref": "chemotion://generic/select_option/draft-01"
        },
        "eln": {
            "type": "object",
            "properties": {
                "version": {
                    "type": "string"
                },
                "base_revision": {
                    "type": "string"
                },
                "current_revision": {
                    "type": "number"
                }
            },
            "required": [
                "version",
                "base_revision",
                "current_revision"
            ]
        }
    },
    "additionalProperties": False,
    "required": [
        "pkg",
        "uuid",
        "klass",
        "layers",
        "select_options"
    ]
}

SchemaRegistry.instance().register(properties_schema)



schema = {
    "$schema": "chemotion:type/research_plan/1.9.3/draft-01/schema#",
    "title": "Research plan Properties",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "description": "The reaction's name"
        },
        "conditions": {
            "type": "string",
            "description": "The reaction's conditions"
        },
        "description": {
            "type": "object",
            "description": "The reaction's description (Quill.js)",
            "properties": {
                "ops": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "insert": {
                                "type": "string"
                            }
                        }
                    }

                }
            }
        },
        "observation": {
            "type": "object",
            "description": "Important observations (Quill.js)",
            "properties": {
                "ops": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "insert": {
                                "type": "string"
                            }
                        }
                    }

                }
            }
        },
        "temperature": {
            "type": "object",
            "description": "A temperature profile of a chemical reaction",
            "properties": {
                "data": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "time": {
                                "type": "string",
                                "pattern": "^\d{2}:\d{2}:\d{2}$",
                            },
                            "value": {
                                "type": "string",
                                "pattern": "^[\d\.]+$",
                            }
                        },
                        "required": ["value", "time"]
                    }
                },
                "userText": {
                    "type": "string"
                },
                "valueUnit": {
                    "type": "string",
                    "enum": ["°C", "°F", "K"],
                }
            },
            "required": ["data", "userText", "valueUnit"]
        },
        "status": {
            "type": "string",
            "enum": ['', 'Planned', 'Running', 'Done', 'Analyses Pending', 'Successful',
                     'Not Successful']
        },
        "purification": {
            "type": ["array", "string"],
            "items": {
                "type": "string",
                "enum": PURIFICATION_OPTIONS
            }
        },
        "duration": {
            "type": "string"
        }
    },
    "required": ["name", "description", "observation",
                 "conditions", "temperature", "status",
                 "purification", "duration"]
}



from enum import Enum

class ListableEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

class SaveActionType(Enum):
    DRAFT = "draft"
    MINOR = "minor"
    MAJOR = "major"

class ElementType(ListableEnum):
    SEGMENT = 'SegmentKlass'
    ELEMENT = 'ElementKlass'
    DATASET = 'DatasetKlass'

class SubFieldType(ListableEnum):
    NUMBER = 'number'
    LABEL = 'label'
    TEXT = 'text'
    SYSTEM_DEFINED = 'system-defined'


class SubTableFieldType(ListableEnum):
    DRAG_MOLECULE = "drag_molecule"
    DRAG_SAMPLE = "drag_sample"
    SELECT = "select"
    SYSTEM_DEFINED = 'system-defined'
    TEXT = 'text'


class FieldType(ListableEnum):
    DUMMY = 'dummy'
    CHECKBOX = "checkbox"
    DATETIME = "datetime"
    DATETIME_RANGE = "datetime-range"
    DRAG_ELEMENT = "drag_element"
    DRAG_MOLECULE = "drag_molecule"
    DRAG_SAMPLE = "drag_sample"
    FORMULA_FIELD = "formula-field"
    INPUT_GROUP = "input-group"
    INTEGER = "integer"
    SELECT = "select"
    SYSTEM_DEFINED = "system-defined"
    TABLE = "table"
    TEXT_FORMULA = 'text-formula'
    TEXT = 'text'
    TEXTAREA = 'textarea'
    UPLOAD = 'upload'


class ConditionOperator(ListableEnum):
    MATCH_ALL = 9
    MATCH_ONE = 1
    MATCH_NONE = 0


class FieldUnit(ListableEnum):
    ACCELERATION = "acceleration"
    AGITATION = "agitation"
    AMOUNT_ENZYME = "amount_enzyme"
    AMOUNT_SUBSTANCE = "amount_substance"
    MOLARITY = "molarity"
    CHEM_DISTANCES = "chem_distances"
    CONCENTRATION = "concentration"
    CONDUCTIVITY = "conductivity"
    CURRENT = "current"
    DEGREE = "degree"
    DENSITY = "density"
    DURATION = "duration"
    ELASTIC_MODULUS = "elastic_modulus"
    ELECTRIC_CHARGE_C = "electric_charge_c"
    ELECTRIC_CHARGE_MOL = "electric_charge_mol"
    ELECTRIC_FIELD = "electric_field"
    ENERGY = "energy"
    ENZYME_ACTIVITY = "enzyme_activity"
    FARADAY = "faraday"
    FLOW_RATE = "flow_rate"
    FREQUENCY = "frequency"
    HEATING_RATE = "heating_rate"
    LENGTH = "length"
    MAGNETIC_FLUX_DENSITY = "magnetic_flux_density"
    MASS = "mass"
    MASS_MOLECULE = "mass_molecule"
    MOLECULAR_WEIGHT = "molecular_weight"
    PERCENTAGE = "percentage"
    PRESSURE = "pressure"
    REACTION_RATE = "reaction_rate"
    SPECIFIC_VOLUME = "specific_volume"
    SPEED = "speed"
    SUBATOMIC_LENGTH = "subatomic_length"
    SURFACE = "surface"
    TEMPERATURE = "temperature"
    TURNOVER_NUMBER = "turnover_number"
    VISCOSITY = "viscosity"
    KINEMATIC_VISCOSITY = "kinematic_viscosity"
    VOLTAGE = "voltage"
    VOLUMES = "volumes"


class LayerColor(ListableEnum):
    NONE = "none"
    BLUE = "primary"
    BLUE_LIGHT = "info"
    GREEN = "success"
    GREY = "default"
    RED = "danger"
    ORANGE = "warning"

UNITS = {
    FieldUnit.ACCELERATION: ["mm_s2"],
    FieldUnit.AGITATION: ["rpm"],
    FieldUnit.AMOUNT_ENZYME: ["u", "mu", "kat", "mkat", "µkat", "nkat"],
    FieldUnit.AMOUNT_SUBSTANCE:  ["mol", "mmol", "umol", "nmol", "pmol"],
    FieldUnit.MOLARITY: ["mol_l", "mmol_l", "umol_l", "nmol_l", "pmol_l"],
    FieldUnit.CHEM_DISTANCES: ["angstrom"],
    FieldUnit.CONCENTRATION: ["ng_l", "mg_l", "g_l"],
    FieldUnit.CONDUCTIVITY: ["s_m"],
    FieldUnit.CURRENT: ["A", "mA", "uA", "nA"],
    FieldUnit.DEGREE: ["degree"],
    FieldUnit.DENSITY: ["g_cm3", "kg_l"],
    FieldUnit.DURATION: ["d", "h", "min", "s"],
    FieldUnit.ELASTIC_MODULUS: ["m_pa", "k_pa", "pa"],
    FieldUnit.ELECTRIC_CHARGE_C: ["ec_c"],
    FieldUnit.ELECTRIC_CHARGE_MOL: ["ec_mol"],
    FieldUnit.ELECTRIC_FIELD: ["v_m"],
    FieldUnit.ENERGY: ["eV", "keV", "j", "k_j"],
    FieldUnit.ENZYME_ACTIVITY: ["u_l", "u_ml"],
    FieldUnit.FARADAY: ["faraday"],
    FieldUnit.FLOW_RATE: ["ul_min", "ml_min", "l_m"],
    FieldUnit.FREQUENCY: ["mhz", "hz", "khz"],
    FieldUnit.HEATING_RATE: ["k_min"],
    FieldUnit.LENGTH: ["mm", "cm", "m"],
    FieldUnit.MAGNETIC_FLUX_DENSITY: ["T"],
    FieldUnit.MASS: ["g", "mg", "ug"],
    FieldUnit.MASS_MOLECULE: ["dalton", "kilo_dalton"],
    FieldUnit.MOLECULAR_WEIGHT: ["g_mol"],
    FieldUnit.PERCENTAGE: ["p"],
    FieldUnit.PRESSURE: ["atm", "pa", "torr"],
    FieldUnit.REACTION_RATE: ["mol_lmin", "mol_lsec"],
    FieldUnit.SPECIFIC_VOLUME: ["cm3_g"],
    FieldUnit.SPEED: ["cm_s", "mm_s", "um_m", "nm_m", "cm_h", "mm_h"],
    FieldUnit.SUBATOMIC_LENGTH: ["um", "nm", "pm"],
    FieldUnit.SURFACE: ["a_2", "um_2", "mm_2", "cm_2"],
    FieldUnit.TEMPERATURE: ["C", "F", "K"],
    FieldUnit.TURNOVER_NUMBER: ["1_s", "1_m"],
    FieldUnit.VISCOSITY: ["pas", "mpas"],
    FieldUnit.KINEMATIC_VISCOSITY: ["m2_s"],
    FieldUnit.VOLTAGE: ["mv", "v"],
    FieldUnit.VOLUMES: ["l", "ml", "ul", "nl"],
}

def get_unit(field_unit: FieldUnit) -> list[str]:
    return UNITS[field_unit]



from chemotion_api.utils import FixedDict


class Condition(FixedDict):
    def __init__(self, id: str, field: str, label: str, layer: str, value:str):
        super().__init__({
            "id": id,
            "field": field,
            "label": label,
            "layer": layer,
            "value": value
        })



import json
import os
from typing import List
from uuid import uuid4

import requests

from chemotion_api.connection import Connection


class Profile:
    def __init__(self, session: Connection, title: str, id: str, description: str, identifiers: list[dict],
                 tables: list[dict], data: dict, matchTables: bool | None = None, ols: str | None = None) -> None:
        self._session = session
        self.ols = ols
        self.data = data
        self.title = title
        self.id = id
        self.description = description
        self.tables = tables
        self.identifiers = identifiers
        self.matchTables = matchTables
        self._is_saved = False

    def to_dict(self) -> dict:
        result = {
            "ols": self.ols,
            "data": self.data,
            "title": self.title,
            "id": self.id,
            "description": self.description,
            "tables": self.tables,
            "identifiers": self.identifiers,
        }
        if self.matchTables is not None:
            result["matchTables"] = self.matchTables
        if self.ols is not None:
            result["ols"] = self.ols
        return result

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def __str__(self):
        return self.title

    def save(self) -> requests.Response:
        if self._is_saved:
            res = self._session.put(f'/api/v1/converter/profiles/{self.id}', data=self.to_dict())
        else:
            res = self._session.post(f'/api/v1/converter/profiles', data=self.to_dict())
        if res.status_code == 200 or res.status_code == 201:
            self._is_saved = True
        return res

    def delete(self):
        return self._session.delete(f'/api/v1/converter/profiles/{self.id}')


class ProfileManager(object):
    """
    Profile manager class
    """

    def __init__(self, session: Connection):
        super().__init__()
        self._session = session
        self._all_profiles = None

    def get_profile(self, profile_id: str) -> Profile:
        return next((p for p in self.profiles if p.id == profile_id), None)

    @property
    def profiles(self) -> List[Profile]:
        if self._all_profiles is None:
            res = self._session.get('api/v1/converter/profiles')
            if res.status_code != 200:
                raise ConnectionError(res.text)

            self._all_profiles = [Profile(self._session, **x) for x in res.json()['profiles']]
        return self._all_profiles

    def new_profile_from_file(self, file_path: str) -> Profile | None:
        if os.path.exists(file_path):
            try:
                with open(file_path, 'r') as f:
                    pd = json.loads(f.read())
                    while self.get_profile(pd['id']) is not None:
                        pd['id'] = uuid4().__str__()
                    return Profile(self._session, **pd)
            except json.decoder.JSONDecodeError:
                pass
        return None


import json
import os

_SOLVER_LIST = None


def get_solvent_list() -> dict[str, dict[str, str | float]]:
    global _SOLVER_LIST
    if _SOLVER_LIST is None:
        json_path = os.path.join(os.path.dirname(__file__), '../elements/empty_elements/solvents.json')
        if os.path.exists(json_path):
            with open(json_path, 'r') as f:
                _SOLVER_LIST = json.loads(f.read())
        else:
            return {}
    return _SOLVER_LIST



import json
import re
import copy
from enum import Enum
from typing import Literal, Optional, Self

from requests import Response

from chemotion_api.elements.utils.converter import convert_value_to_default
from chemotion_api.elements.utils.equivalent import get_equivalent
from chemotion_api.user import User
from chemotion_api.elements.abstract_element import AbstractElement
from chemotion_api.elements.empty_elements import init_container
from collections.abc import MutableMapping
from chemotion_api.connection import Connection
from chemotion_api.utils import UnitDict, FixedDict, PropertyDict
from chemotion_api.utils.solvent_manager import get_solvent_list
from chemotion_api.elements.utils import GasType, QuantityUnit, calculate_quantity, GasCalculations


class GasPhaseObj(FixedDict):
    default_val = {
        "time": {
            "unit": "h",
            "value": None
        },
        "temperature": {
            "unit": "K",
            "value": None
        },
        "turnover_number": None,
        "part_per_million": None,
        "turnover_frequency": {
            "unit": "TON/h",
            "value": None
        }
    }

    def __init__(self, sample: 'Sample', kwargs: Optional[dict] = None):
        self._sample = sample
        if kwargs is None:
            kwargs = {**self.default_val}
        else:
            kwargs = copy.deepcopy(kwargs)
        kwargs['turnover_frequency'] = UnitDict(**kwargs['turnover_frequency'])
        kwargs['temperature'] = UnitDict(**kwargs['temperature'])
        kwargs['time'] = UnitDict(**kwargs['time'])
        super().__init__(kwargs)

    @property
    def ppm(self) -> float:
        return self.data['part_per_million']

    @ppm.setter
    def ppm(self, value: float):
        self['part_per_million'] = value

    @property
    def time_in_h(self) -> float:
        return convert_value_to_default(**self.data['time'])

    @time_in_h.setter
    def time_in_h(self, value: float):
        self['time'] = UnitDict(unit='h', value=value)

    def set_time(self, value: float, unit: str = 'h'):
        self['time'] = UnitDict(unit='K', value=convert_value_to_default(value=value, unit=unit))

    @property
    def temperature_in_k(self) -> float:
        return convert_value_to_default(**self.data['temperature'])

    @temperature_in_k.setter
    def temperature_in_k(self, value: float):
        self['temperature'] = UnitDict(unit='K', value=value)

    def set_temperature(self, value: float, unit: str = 'K'):
        self['temperature'] = UnitDict(unit='K', value=convert_value_to_default(value=value, unit=unit))

    def to_dict(self):
        return {
            **self.data,
            'turnover_frequency': self.data['turnover_frequency'].to_dict(),
            'temperature': self.data['temperature'].to_dict(),
            'time': self.data['time'].to_dict()
        }

    def copy_with_sample(self, sample: 'Sample') -> Self:
        return self.__class__(sample, self.to_dict())


class SampleType(Enum):
    SAMPLE_TYPE_MIXTURE = 'Mixture'
    SAMPLE_TYPE_MICROMOLECULE = 'Micromolecule'


class SampleVariationType(Enum):
    STARTING_MATERIAL = 'startingMaterials'
    REACTANT = 'reactants'
    PRODUCT = 'products'
    SOLVENTS = 'solvents'

    @classmethod
    def from_role(cls, sample: 'Sample') -> Self:
        role = sample.role_in_reaction
        return cls.from_role_name(role)

    @classmethod
    def from_role_name(cls, role: str) -> Self:
        if role == 'starting_materials':
            return SampleVariationType.STARTING_MATERIAL
        return cls(role)


class SolventList(list):
    """
    In generale it contains a list of molecules. It allows to add molecules from the solvents list.
    """

    def __init__(self, session: Connection, *args):
        if args is None or args[0] is None:
            args = []
        super().__init__(*args)
        self._session = session

    def add_new_smiles(self, smiles: str):
        """
        Add a solvent by smiles code

        :param smiles: Smiles code
        """

        m = MoleculeManager(self._session).create_molecule_by_smiles(smiles)
        self.append({
            "label": m.get('iupac_name', m.get('sum_formular')),
            "smiles": smiles,
            "inchikey": m.get("inchikey"),
            "ratio": 1
        })

    def add_new_name(self, name: str):
        """
        Add a solvent from the solvent list by name

        :param name: Solvent Name
        """

        solvent_info = get_solvent_list().get(name)
        if solvent_info is None:
            raise KeyError(
                'Solver: "{}" is not available. Run instance.get_solvent_list() to see all valid solver names'.format(
                    name))

        m = MoleculeManager(self._session).create_molecule_by_smiles(solvent_info['smiles'])
        self.append({
            "label": name,
            "smiles": m.get('cano_smiles'),
            "inchikey": m.get("inchikey"),
            "ratio": 1
        })


class Molecule(MutableMapping):
    """
    Contains the Molecule information. This class extents MutableMapping and can be used as normal dict object.
    However, it ensures that the following keys are set and used if the sample which contains the molecule is saved.

    :key: "boiling_point" {float}
    :key: "cano_smiles" {str}
    :key: "density" {float}
    :key: "inchikey" {str}
    :key: "inchistring" {str}
    :key: "melting_point" {float}
    """

    def __init__(self, data):
        self.all_store = dict(data)
        self.id = data.get('id')
        for key in ["boiling_point", "molecular_weight", "exact_molecular_weight", "cano_smiles", "density", "inchikey",
                    "inchistring", "melting_point", "sum_formular"]:
            self.all_store[key] = self.all_store.get(key)

    def __getitem__(self, key: str):
        key = self._keytransform(key)
        return self.all_store[key]

    def __setitem__(self, key: str, value):
        key = self._keytransform(key)
        self.all_store[key] = value

    def __delitem__(self, key):
        del self.all_store[self._keytransform(key)]

    def __iter__(self):
        return iter(self.all_store)

    def __len__(self):
        return len(self.all_store)

    def _keytransform(self, key):
        return key

    def clean(self) -> dict:
        return {
            "density": self.all_store.get("density"),
            "melting_point": self.all_store.get("melting_point"),
            "boiling_point": self.all_store.get("boiling_point"),
            "inchistring": self.all_store.get("inchistring"),
            "cano_smiles": self.all_store.get("cano_smiles"),
            "inchikey": self.all_store.get("inchikey")
        }


class MoleculeManager:
    """
    The MoleculeManager can be used to load and crate molecules in teh Chemotion Instance
    """

    def __init__(self, session: Connection):
        self._session = session

    def create_molecule_by_smiles(self, smiles_code: str, density: float = 0.0) -> Molecule:
        """
        Creates a new molecule for a given smiles code

        :param smiles_code: Simles code of te sample
        :return: Molecule obj
        """
        smiles_url = "/api/v1/molecules/smiles"
        payload = {
            "editor": "ketcher",
            "smiles": smiles_code
        }
        res = self._session.post(smiles_url,
                                 data=json.dumps(payload))
        if res.status_code != 201:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        res_json = res.json()
        res_json['density'] = density
        return Molecule(res_json)

    def create_molecule_by_cls(self, host_url, session, inchikey) -> Molecule:
        """Not implemented yet!!"""
        raise NotImplementedError


class Sample(AbstractElement):
    """
    A chemotion Sample object.
    It extends the :class:`chemotion_api.elements.abstract_element.AbstractElement`

    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the sample with ID 1
    >>> s = instance.get_sample(1)
    >>> # Set the real amount to 3.7 g
    >>> s.set_amount(3.7, "g")
    >>> # Set the boiling point range from 40 to 50
    >>> s.properties["boiling_point_lowerbound"] = 40
    >>> s.properties["boiling_point_upperbound"] = 50
    >>> # Set the external label to "Sample_X334D"
    >>> s.properties["external_label"] = "Sample_X334D"
    >>> # Save the structure Image
    >>> with open("./sample_structure.svg", "wb+") as f:
    >>>     f.write(s.load_image().content)
    >>> # Save the sample
    >>> s.save()
    """

    material_types = ['starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']

    def __init__(self, *args, **kwargs):
        self._molecule = kwargs.pop('molecule', None)
        self.is_split = False
        self._reaction = None
        self._role_in_reaction = None
        self._is_solvent = kwargs.pop("is_solvent", False)
        super().__init__(*args, **kwargs)

    def copy_params(self):
        pms = super().copy_params()
        pms["is_solvent"] = self.role_in_reaction == 'solvents'
        pms["molecule"] = self._molecule
        return pms

    def _set_json_data(self, json_data):
        super()._set_json_data(json_data)
        self._molecule = Molecule(json_data.get('molecule'))
        self._svg_file = json_data.get('sample_svg_file')
        self.is_split = json_data.get('is_split', False)
        self._children_count = json_data.get('children_count', )

    @property
    def reaction(self):
        if self._reaction is None:
            try:
                self._reaction = self.reload_reaction()
            except KeyError:
                return None
        return self._reaction

    @reaction.setter
    def reaction(self, reaction):
        self._role_in_reaction = None
        self._reaction = reaction

    def reload_reaction(self):
        from chemotion_api import Instance
        try:
            return Instance(self._session).get_reaction(self.json_data['tag']['taggable_data']['reaction_id'])
        except (KeyError, IndexError):
            raise KeyError(f'Could not find Reaction of Sample: {self}')

    @property
    def role_in_reaction(self) -> Literal[
        'starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']:
        if self._role_in_reaction is None:
            self._role_in_reaction = self.reload_role_in_reaction()
        return self._role_in_reaction

    @role_in_reaction.setter
    def role_in_reaction(self, role_in_reaction: Literal[
        'starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']):
        self._role_in_reaction = role_in_reaction
        if role_in_reaction not in self.material_types:
            raise ValueError(f'Unknown role: {role_in_reaction}')
        self._role_in_reaction = role_in_reaction

    def reload_role_in_reaction(self) -> Optional[Literal[
        'starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']]:
        if self.reaction is None:
            return None

        element: Literal['starting_materials', 'reactants', 'products', 'solvents', 'purification_solvents']
        for element in self.material_types:
            mat_list = self.reaction.properties[element]
            for mat in mat_list:
                if mat.id == self.id:
                    return element
        raise KeyError(f'Could not find Role In {self}')

    @property
    def properties(self) -> PropertyDict:
        """
        The properties property contains all data which can be altered
        through the chemotion api from the main tab of the sample.


        :key solvent: {:class:`chemotion_api.elements.abstract_element.SolventList`}
        :key description: {str}
        :key external_label: {str}
        :key boiling_point_lowerbound: {float}
        :key boiling_point_upperbound: {float}
        :key melting_point_lowerbound: {float}
        :key melting_point_upperbound: {float}
        :key target_amount: {dict} has a unit {str} ('l', 'g' or 'mol') and an amount {float}. The amount of the sample used as starting material or reactant.
        :key molarity: {dict} with value: float and unit: str
        :key real_amount: {dict} has a unit {str} ('l', 'g' or 'mol') and an amount {float}. The amount of sample obtained from a reaction.
        :key stereo: {str}
        :key location: {str}
        :key is_top_secret: {bool}
        :key is_restricted: {bool}
        :key purity: {float}
        :key density: {float}
        :key user_labels: {list}
        :key decoupled: {bool}
        :key waste: {bool}
        :key metrics: {str}
        :key sum_formula: {str}
        :key equivalent: {str}
        :key coefficient: {str}
        :key reaction_description: {str}
        :key gas_type {:class:`chemotion_api.elements.sample.GasType`}
        :key inventory_sample: {bool}
        :key gas_phase_data: {dict}

        Readonly properties:

        :key name: {str, readonly}
        :key short_label: {str, readonly}

        :return: Element properties
        """
        return super().properties

    @property
    def molecule(self) -> Molecule:
        """
        The molecule of the sample
        """
        if self._molecule is None:
            return Molecule({})
        return self._molecule

    @molecule.setter
    def molecule(self, molecule: Molecule):
        """
        The molecule of the sample
        """

        self._molecule = molecule

    @property
    def vessel_size(self):
        if self.reaction is None:
            return None
        return self.reaction.vessel_size_in_l

    @property
    def gram(self):
        return self._get_converted_amount(QuantityUnit.Gram)

    @property
    def mol(self):
        return self._get_converted_amount(QuantityUnit.Mole)

    @property
    def liter(self):
        return self._get_converted_amount(QuantityUnit.Liter)

    @property
    def ton(self):
        if self.gas_type != GasType.GAS or self.reaction is None or  self.reaction.get_catalyst() is None:
            return None
        return self.mol / self.reaction.get_catalyst().mol

    @property
    def ton_per_h(self):
        ton = self.ton
        if ton is None:
            return None
        try:
            return ton / self.properties['gas_phase_data'].time_in_h
        except TypeError or ZeroDivisionError:
            return 0

    def _get_converted_amount(self, output_unit):
        key = 'real_amount' if self.role_in_reaction == 'products' else 'target_amount'
        gp: GasPhaseObj = self.properties.get('gas_phase_data')
        if gp is not None:
            gas_attres = {'temperature_in_k': gp.temperature_in_k, 'ppm': gp.ppm, 'time_in_h': gp.time_in_h}

        else:
            gas_attres = {}
        return calculate_quantity(
            current_value=convert_value_to_default(**self.properties[key]),
            current_unit=QuantityUnit.from_string(self.properties[key]['unit']),
            output_unit=output_unit,
            molecular_weight=self.molecule['molecular_weight'],
            purity=self.properties['purity'],
            molarity=self.properties['molarity']['value'],
            density=self.properties['density'],
            gas_type=self.gas_type,
            vessel_size=self.vessel_size,
            **gas_attres
        )

    @property
    def equivalent(self) -> float:
        return self.properties['equivalent']

    @equivalent.setter
    def equivalent(self, equivalent: float):
        self.properties['equivalent'] = equivalent

    @property
    def gas_type(self) -> GasType:
        gt = self._properties.get('gas_type', GasType.OFF)
        if self.reaction is None and gt is not GasType.OFF:
            self.set_gas_type(GasType.OFF)
            return GasType.OFF
        return gt

    @property
    def reference(self) -> bool:
        return self.properties['reference']

    @reference.setter
    def reference(self, ref: bool):
        self.set_ref(ref)

    @gas_type.setter
    def gas_type(self, gas_type: GasType | str):
        self.set_gas_type(gas_type)

    def set_gas_type(self, gas_type: GasType | str, reaction_update: bool = True):
        if isinstance(gas_type, str):
            gas_type = GasType(gas_type)

        if gas_type != GasType.OFF and self.reaction is not None and not self.reaction.gas_schema:
            raise ValueError(f'Reaction is not a gas schema reaction')
        if gas_type not in self.allowed_gas_types():
            raise ValueError(f'Reaction is not a gas schema reaction')
        has_change = self._properties['gas_type'] != gas_type and self.reaction is not None
        if has_change and gas_type in [GasType.FEEDSTOCK, GasType.CATALYST]:
            gas_elem = self.reaction.get_gas_component(gas_type)
            if gas_elem is not None:
                gas_elem.set_gas_type(GasType.OFF, False)
        origin_gas_type = self.gas_type
        self._properties['gas_type'] = gas_type
        if has_change and gas_type == GasType.GAS:
            self.properties['gas_phase_data'] = GasPhaseObj(self)
        if has_change and origin_gas_type == GasType.FEEDSTOCK:
            cat_sample = self.reaction.get_catalyst()
            if cat_sample is not None:
                cat_sample.gas_type = GasType.OFF
        if has_change and reaction_update:
            self.reaction.update_amount()

    def load_image(self) -> Response:
        """
        Loads the sample structure as svg image

        :return: Response with the svg as content
        """

        image_url = "/images/samples/{}".format(self._svg_file)
        res = self._session.get(image_url)
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        return res

    def split(self):
        """
        Splits the sample.

        :return: A new sample object
        :rtype: :class:`chemotion_api.elements.sample.Sample`
        """
        self.save(True)
        split_sample = copy.deepcopy(self.json_data)
        split_sample['parent_id'] = self.id
        split_sample['id'] = None

        if 'tag' in split_sample:
            del split_sample['tag']
        split_sample['created_at'] = None
        split_sample['updated_at'] = None
        split_sample['target_amount_value'] = 0
        split_sample['real_amount_value'] = None
        split_sample['is_split'] = True
        split_sample['is_new'] = True
        split_sample['short_label'] += '-NaN'

        split_sample['container'] = init_container()
        return Sample(generic_segments=self._generic_segments, session=self._session, json_data=split_sample)

    def copy(self):
        """
        Splits the sample.

        :return: A new sample object
        :rtype: :class:`chemotion_api.elements.sample.Sample`
        """
        self.save()
        split_sample = copy.deepcopy(self.json_data)
        split_sample['id'] = None

        if 'tag' in split_sample:
            del split_sample['tag']
        split_sample['created_at'] = None
        split_sample['updated_at'] = None
        split_sample['is_split'] = False
        split_sample['is_new'] = True
        split_sample['short_label'] = User.load_me(self._session).get_next_short_label('sample')

        split_sample['container'] = init_container()
        return Sample(generic_segments=self._generic_segments, session=self._session, json_data=split_sample)

    def toggle_decoupled(self):
        """Decoupls the sample from the molecule"""
        self.properties['decoupled'] = not self.properties['decoupled']

    def clean_data(self, *args, **kwargs):
        serialize_data = super().clean_data(*args, **kwargs)
        if self._is_solvent:
            keys = ["real_amount_value", "gas_phase_data", "name", "boiling_point_lowerbound",
                    "boiling_point_upperbound", "melting_point_lowerbound",
                    "melting_point_upperbound", "dry_solvent", "equivalent", "sample_svg_file", "parent_id"]
            for key in keys:
                if key in serialize_data:
                    del serialize_data[key]
        return serialize_data

    def _parse_properties(self) -> dict:
        melting_range = re.split(r'\.{2,3}', self.json_data.get('melting_point')) if self.json_data.get(
            'melting_point') is not None else ['', '']
        boiling_range = re.split(r'\.{2,3}', self.json_data.get('boiling_point')) if self.json_data.get(
            'boiling_point') is not None else ['', '']

        gas_props = {}
        if self.json_data.get('gas_type') is not None:
            gas_props['gas_type'] = GasType(self.json_data.get('gas_type'))
        else:
            gas_props['gas_type'] = GasType.OFF

        if self.json_data.get('gas_phase_data') is not None:
            gas_props['gas_phase_data'] = GasPhaseObj(self, self.json_data.get('gas_phase_data'))

        return {
            'name': self.json_data.get('name'),
            'short_label': self.json_data.get('short_label'),
            'solvent': SolventList(self._session, self.json_data.get('solvent', [])),
            'description': self.json_data.get('description'),
            'external_label': self.json_data.get('external_label'),
            'boiling_point_lowerbound': int(boiling_range[0]) if boiling_range[0].isdigit() else None,
            'boiling_point_upperbound': int(boiling_range[1]) if boiling_range[1].isdigit() else None,
            'melting_point_lowerbound': int(melting_range[0]) if melting_range[0].isdigit() else None,
            'melting_point_upperbound': int(melting_range[1]) if melting_range[1].isdigit() else None,
            'target_amount': UnitDict(unit=self.json_data.get('target_amount_unit'),
                                      value=self.json_data.get('target_amount_value')),
            'real_amount': UnitDict(unit=self.json_data.get('real_amount_unit'),
                                    value=self.json_data.get('real_amount_value')),
            'stereo': self.json_data.get('stereo'),
            'location': self.json_data.get('location'),
            'is_top_secret': self.json_data.get('is_top_secret'),
            'is_restricted': self.json_data.get('is_restricted'),
            'purity': self.json_data.get('purity'),
            'density': self.json_data.get('density'),
            'user_labels': self.json_data.get('user_labels', []),
            'decoupled': self.json_data.get('decoupled'),
            'waste': self.json_data.get('waste', False),
            'reference': self.json_data.get('reference'),
            'metrics': self.json_data.get('metrics'),
            'sum_formula': self.json_data.get('sum_formula'),
            'equivalent': self.json_data.get('equivalent'),
            'coefficient': self.json_data.get('coefficient', 1),
            'reaction_description': self.json_data.get('reaction_description'),
            'molarity': UnitDict(unit=format(self.json_data.get('molarity_unit')),
                                 value=self.json_data.get('molarity_value')),

            'molecular_mass': self.json_data.get('molecular_mass'),
            'inventory_sample': self.json_data.get('inventory_sample'),
            'position': self.json_data.get('position'),
            'show_label': self.json_data.get('show_label')

        } | gas_props

    def _clean_properties_data(self, serialize_data: dict | None = None) -> dict:
        if serialize_data is None:
            serialize_data = {}

        keys = ["name", "boiling_point_lowerbound", "boiling_point_upperbound", "melting_point_lowerbound",
                "melting_point_upperbound", "dry_solvent", "equivalent", "sample_svg_file", "parent_id"]

        for key in keys:
            self._set_serialized_if_exist(key, serialize_data)

        serialize_data['description'] = self.properties.get('description')
        serialize_data['external_label'] = self.properties.get('external_label')
        serialize_data['short_label'] = self.properties.get('short_label')
        serialize_data['solvent'] = self.properties.get('solvent')
        serialize_data['stereo'] = self.properties.get('stereo').to_dict()
        serialize_data['location'] = self.properties.get('location')
        serialize_data['purity'] = self.properties.get('purity')
        serialize_data['user_labels'] = self.properties.get('user_labels')
        serialize_data['decoupled'] = self.properties.get('decoupled')
        serialize_data['density'] = self.properties.get('density')
        serialize_data['metrics'] = self.properties.get('metrics')

        serialize_data['is_top_secret'] = self.properties.get('is_top_secret')
        serialize_data['molarity_unit'] = self.properties.get('molarity').get('unit')
        serialize_data['molarity_value'] = self.properties.get('molarity').get('value')
        serialize_data['molecular_mass'] = self.properties.get('molecular_mass')

        serialize_data['waste'] = self.properties.get('waste')
        serialize_data['reference'] = self.properties.get('reference', False)
        serialize_data['sum_formula'] = self.properties.get('sum_formula')
        serialize_data['coefficient'] = self.properties.get('coefficient', 1)

        serialize_data['target_amount_unit'] = self.properties.get('target_amount').get('unit')
        serialize_data['target_amount_value'] = self.properties.get('target_amount').get('value')

        serialize_data['real_amount_unit'] = self.properties.get('real_amount').get('unit')
        serialize_data['real_amount_value'] = self.properties.get('real_amount').get('value')

        serialize_data['is_split'] = self.is_split

        serialize_data['molfile'] = self.json_data.get('molfile')
        serialize_data['inventory_sample'] = self.properties.get('inventory_sample')
        serialize_data['gas_type'] = self.properties.get('gas_type').value
        if self.gas_type == GasType.GAS and self.properties.get('gas_phase_data') is not None:
            serialize_data['gas_phase_data'] = self.properties.get('gas_phase_data').to_dict()
        else:
            serialize_data['gas_phase_data'] = None

        serialize_data['dry_solvent'] = self.json_data.get('dry_solvent')
        serialize_data['residues'] = self.json_data.get('residues')
        serialize_data['imported_readout'] = self.json_data.get('imported_readout')
        serialize_data['xref'] = self.json_data.get('xref')
        serialize_data['elemental_compositions'] = self.json_data.get('elemental_compositions')

        self._set_serialized_if_exist('molecule_name_id', serialize_data, self.json_data)
        self._set_serialized_if_exist('collection_id', serialize_data, self.json_data)
        self._set_serialized_if_exist('position', serialize_data)
        self._set_serialized_if_exist('show_label', serialize_data)
        self._set_serialized_if_exist('components', serialize_data)
        self._set_serialized_if_exist('sample_type', serialize_data)

        if self.molecule is not None and self.molecule.id is not None:
            serialize_data['molecule'] = self.molecule.clean()
            serialize_data['molecule_id'] = self.molecule.id
            serialize_data['molfile'] = self.molecule['molfile']

        return serialize_data

    def _set_serialized_if_exist(self, key: str, serialize_data: dict, origen_data: Optional[dict] = None):
        if origen_data is None:
            value = self.properties.get(key, self.json_data.get(key, '____NOT_EXISTING____'))
        else:
            value = origen_data.get(key, '____NOT_EXISTING____')
        if value != '____NOT_EXISTING____':
            serialize_data[key] = value

    def allowed_gas_types(self):
        if self.reaction is None:
            return [GasType.OFF]
        if self.role_in_reaction in ['starting_materials', 'reactants']:
            feedstock_mat = self.reaction.get_feedstock()
            if feedstock_mat is not None:
                return [GasType.OFF, GasType.FEEDSTOCK, GasType.CATALYST]
            else:
                return [GasType.OFF, GasType.FEEDSTOCK]
        if self.role_in_reaction == 'products':
            return [GasType.OFF, GasType.GAS]
        return [GasType.OFF]

    def allowed_quantity_units(self):
        return self._allowed_quantity_units()

    def _allowed_quantity_units(self):
        if self.role_in_reaction == 'solvents':
            return [QuantityUnit.Liter]
        elif self.role_in_reaction == 'purification_solvents':
            return [QuantityUnit.Liter]
        else:
            if self.gas_type == GasType.GAS:
                return []
            if self.gas_type == GasType.FEEDSTOCK:
                default_set = [QuantityUnit.Liter]
            else:
                default_set = [QuantityUnit.Gram]
                if self.properties['density'] != 0 or self.properties['molarity']['value'] != 0:
                    default_set += [QuantityUnit.Liter]

            if self.role_in_reaction in ['starting_materials', 'reactants']:
                return default_set + [QuantityUnit.Mole, QuantityUnit.Equivalent]
            elif self.role_in_reaction == 'products':
                return default_set + []
        return default_set + [QuantityUnit.Mole]

    def update_equivalent(self):
        try:
            self._set_equivalent()
        except KeyError:
            self._properties['equivalent'] = None

    def update_amount(self):
        amount = self.get_amount()
        unit = QuantityUnit.from_string(amount['unit'])
        allowed_units = self.allowed_quantity_units()
        if unit not in allowed_units and allowed_units:
            self.set_amount(0.0, allowed_units[-1])

    def _set_equivalent(self, value: Optional[float] = None):
        unit = QuantityUnit.Equivalent
        if value is not None:
            if unit not in self.allowed_quantity_units():
                raise KeyError(
                    f'The value {unit.value}  cannot be set for this sample. This is most likely due to the role of this sample in the reaction.')
            self._properties['equivalent'] = value
            self.set_amount(self.reaction.reference_sample.mol * value, QuantityUnit.Mole)
        elif self.properties['reference']:
            self._properties['equivalent'] = 1
        else:
            try:
                ref_mol = self.reaction.reference_sample.mol
            except AttributeError:
                ref_mol = None
            try:
                ppm = self.properties.get('gas_phase_data').ppm
                temperature_in_k = self.properties.get('gas_phase_data').temperature_in_k
            except AttributeError:
                ppm = None
                temperature_in_k = None
            self._properties['equivalent'] = get_equivalent(self, self.mol, ref_mol, ppm=ppm,
                                                            temperature_in_k=temperature_in_k)

    def set_ref(self, is_ref: bool, check_ref: bool = True):
        if check_ref:
            self.properties['reference'] = is_ref
        else:
            self._properties['reference'] = is_ref

    def get_amount(self):
        key = 'real_amount' if self.role_in_reaction == 'products' else 'target_amount'
        return self._properties[key]

    def set_amount(self, value: float, unit: Optional[QuantityUnit | str] = None):
        if isinstance(unit, str):
            unit_str = unit
            unit = QuantityUnit.from_string(unit_str)
            if unit is None:
                raise ValueError(f'{unit_str} is not a valid unit')
        amount_dict = ['target_amount', 'real_amount']
        [key, zero_key] = amount_dict.__reversed__() if self.role_in_reaction == 'products' else amount_dict
        if unit is None:
            unit = QuantityUnit.from_string(self.properties[key]['unit'])

        if unit not in self.allowed_quantity_units():
            raise ValueError(f'This sample can set the {unit.value}')
        if value is None:
            value = 0.0
        if value < 0:
            raise ValueError(f'The amount must be >= 0 {unit.value}')
        if unit == QuantityUnit.Yield:
            raise KeyError('Yield cannot be set!!')
        elif unit == QuantityUnit.Equivalent:
            self._set_equivalent(value)
        else:
            self._properties[key] = UnitDict(unit=str(unit.value), value=value)
            self._properties[zero_key] = UnitDict(unit=str(unit.value), value=None)
            try:
                self.reaction.update_equivalent()
            except (KeyError, ValueError, TypeError, ZeroDivisionError, AttributeError):
                pass

    def on_properties_change(self, dict_obj, name, value):
        try:
            if name in ['target_amount', 'real_amount']:
                if len(name) > 1:
                    dict_copy = {**dict_obj[-1], name[-1]: value[-1]}
                    self.set_amount(**dict_copy)
                return False
        except ValueError:
            pass
        if name[-1] == 'vessel_size':
            self._properties[name[-1]] = value[-1]
            return False
        if name[0] == 'gas_phase_data':
            dict_obj[-1][name[-1]] = value[-1]
            self.update_equivalent()
            return False
        if name[-1] == 'gas_type':
            self.gas_type = value[-1]
            return False

        if name[-1] == 'purity':
            if value[-1] <= 0 or value[-1] > 1:
                raise ValueError('The purity must be >= 0 and <= 1')
            else:
                self._properties[name[-1]] = float(value[-1])
                self.reaction.update_equivalent()
                return False
        if name[-1] == 'coefficient':
            if value[-1] <= 0:
                raise ValueError('The coefficient must be > 0')
            else:
                self._properties[name[-1]] = int(value[-1])
                self.reaction.update_equivalent()
                return False
        if name[-1] == 'reference':
            if value[-1]:
                self.reaction.reference_sample = self
            else:
                self.reaction.reference_sample = None

        unit = QuantityUnit.from_string(name[-1])
        if unit == QuantityUnit.Yield:
            self.set_amount(value[-1], unit)
            return False
        if unit == QuantityUnit.Equivalent:
            dict_obj[-1]['equivalent'] = 0
            self.set_amount(value[-1], unit)
            return False

        return True



generic_element_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "id": {
            "type": "number"
        },
        "uuid": {
            "type": ["string", "null"]
        },
        "label": {
            "type": "string"
        },
        "desc": {
            "type": "string"
        },
        "properties_template": {
            "$ref": "chemotion://generic/properties/draft-01"
        },
        "properties_release": {
            "$ref": "chemotion://generic/properties/draft-01"
        },
        "is_active": {
            "type": "boolean"
        },
        "version": {
            "type": ["null", "string"]
        },
        "place": {
            "type": "number"
        },
        "released_at": {
            "type": "string"
        },
        "identifier": {
            "type": ["string", "null"]
        },
        "sync_time": {},
        "created_by": {
            "type": "number"
        },
        "updated_by": {
            "type": "number"
        },
        "created_at": {
            "type": "string"
        },
        "updated_at": {
            "type": "string"
        },
    },
    "additionalProperties": False,
    "required": [
        "id",
        "uuid",
        "label",
        "desc",
        "properties_template",
        "properties_release",
        "is_active",
        "version",
        "place",
        "released_at",
        "identifier",
        "sync_time",
        "created_at",
        "updated_at",
    ]
}


import yaml

def read_docker_compose(file_path: str) -> dict[str:dict|str]:
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)


def write_docker_compose(file_path: str, content: dict[str:dict|str]):
    with open(file_path, 'w') as file:
        yaml.dump(content, file, default_flow_style=False, sort_keys=False)


def update_images(file_path: str, **kwargs):
    content = read_docker_compose(file_path)

    for service, image in kwargs.items():
        if service in content['services']:
            content['services'][service]['image'] = image

    write_docker_compose(file_path, content)


def update_eln_port(file_path: str, eln_port: int):
    content = read_docker_compose(file_path)
    content['services']['eln']['ports'] = [f'{eln_port}:4000']
    write_docker_compose(file_path, content)


def update_db_port(file_path: str, postgresql_port: int):
    content = read_docker_compose(file_path)
    content['services']['db']['ports'] = [f'{postgresql_port}:5432']
    write_docker_compose(file_path, content)


def get_eln_port(file_path: str):
    content = read_docker_compose(file_path)
    return int(content['services']['eln']['ports'][0].split(':')[0])


def get_db_port(file_path: str):
    content = read_docker_compose(file_path)
    return int(content['services']['db']['ports'][0].split(':')[0])


from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

generic_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/cond_field/draft-01",
    "title": "Schema for generic condition field",
    "type": "object",
    "properties": {
        "id": {
            "type": "string"
        },
        "field": {
            "type": "string"
        },
        "label": {
            "type": "string"
        },
        "layer": {
            "type": "string"
        },
        "value": {
            "type": "string"
        }
    },
    "additionalProperties": False,
    "required": [
        "id",
        "field",
        "label",
        "layer",
        "value"
    ]
}

SchemaRegistry.instance().register(generic_schema)


import typing
from typing import TypeVar
from chemotion_api.collection import RootCollection, Collection
from chemotion_api.converter import ProfileManager
from chemotion_api.element_manager import ElementManager
from chemotion_api.inbox import Inbox
from chemotion_api.labImotion import GenericManager
from chemotion_api.search.reaction import ReactionSearcher
from chemotion_api.search.sample import SampleSearcher
from chemotion_api.user import User, Admin, Person, Device, Group, DeviceManager
from urllib.parse import urlparse
from chemotion_api.elements import ElementSet, Wellplate, Sample, Reaction, GenericElement, ResearchPlan, \
    AbstractElement
from chemotion_api.elements.sample import MoleculeManager
from chemotion_api.connection import Connection

TInstance = TypeVar("TInstance", bound="Instance")


class Instance:
    """
    The instance object is the core object of the Chemotion API. In order for the API to work,
    a connection to a chemotion must first be established.
    an Instance object manges such a connection. To initializes an instance it needs
    the host URL of the chemotion server as a string.

    :param host_url: URL for the new :class:`Request` object

    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")


    """

    def __init__(self, host_url: str | Connection, verify_ssl: bool = True):
        if isinstance(host_url, Connection):
            self._con = host_url
        else:
            self._con = Connection(host_url, verify_ssl)
        self._root_col = None
        self._inbox = None
        self.element_manager = ElementManager(self._con)

    @property
    def host_url(self) -> str:
        """
        The url as string of the instance

        :return: host URL
        """
        return self._con.host_url

    @property
    def token(self) -> str:
        """
        The token property returns the JWT (java web token)

        :return: JWT
        """

        return self._con.bearer_token

    @property
    def device_manager(self) -> DeviceManager:
        """
        A DeviceManager instance

        :return: DeviceManager
        """
        return DeviceManager(self._con)

    @property
    def inbox(self) -> Inbox:
        """
        Get the data file inbox

        :return: Inbox Object
        """
        if self._inbox is None:
            self._inbox = Inbox(self._con)
        return self._inbox

    def test_connection(self) -> TInstance:
        """
        This test_connection methode simply test if the connection to a chemotion instance can be established.
        The instance does not need to be logged in to use this methode.

        :return: the instance self

        :raises ConnectionError: if the connection cannot be established.
        """
        ping_url = "/api/v1/public/ping"
        res = self._con.get(url_path=ping_url)
        if res.status_code != 204:
            raise ConnectionError('Could not ping the Chemotion instance: {}'.format(self.host_url))
        return self

    def get_all_types(self) -> list[str]:
        """
        Generates a list af all JSON LD types in the system.
        :return: List of JSON LD types
        """
        return self.element_manager.all_types

    def login(self, user: str, password: str) -> TInstance:
        """
        This login methode allows you to log in to a chemotion instance.
         To do this, you need a valid chemotion user (abbreviation name or e-mail) and password.

        :param user: abbreviation name or e-mail of a chemotion user
        :param password: The password of the user

        :return: the instance itself

        :raise ConnectionError: if the logging was not successful.
        """

        payload = {'username': user, 'password': password}
        login_url = "/api/v1/public/token"

        res = self._con.post(url_path=login_url,
                             headers=self._con.get_default_session_header(),
                             data=payload)

        if res.status_code != 201:
            raise PermissionError('Could not login!!')
        self._con.set_bearer_token(res.json()['token'])
        return self

    def login_token(self, jwt_token: str) -> TInstance:
        """
        You can use this login_token method to log in to a Chemotion instance.
        To do this, you need a valid Chemotion jwt token. This token can be generated using the
        get_token method.

        :param jwt_token: a Java Web Token.

        :return: the instance itself

        :raise ConnectionError: if the logging was not successful.
        """

        self._con.set_bearer_token(jwt_token)
        return self

    def get_profile_manager(self):
        return ProfileManager(self._con)

    def get_user(self) -> Admin | Person | Group | Device:
        """
        This get_user methode initializes a new User object. The user object allows you to read and edit your user data.

        :return: a new User instance

        :raise PermissionError: if the userdata cannot be fetched. Make sure that you are logged in.
        """
        u = User.load_me(self._con)
        return u

    def find_user_by_name(self, name: str) -> list[Person]:
        """
        This find_user_by_name methode finds all user by name and returns the three top matches.

        :param name: Potential name or name snippet of a user
        :return: list u matching user

        :raise PermissionError: if the userdata cannot be fetched. Make sure that you are logged in.
        """

        return self.get_user().find_user_by_name(self._con, name)

    def get_root_collection(self, reload=True) -> RootCollection:
        """
        The root collection can be compared to the all collection in the web client of Chemotion.
        Within this python client it is needed to navigate through all collections. A collection is foundtion of fetching
        and creating new elements.

        :param reload: If ture caned information are ignored
        :return:
        """
        if reload or self._root_col is None:
            self._root_col = RootCollection(self._con)
            self._root_col.set_element_manager(self.element_manager)
            self._root_col.load_collection()
            self._root_col.load_sync_collection()
        return self._root_col

    def get_collection_by_id(self, collection_id: int) -> Collection:
        """
        Find a collection by its id.
        Within this python client it is needed to navigate through all sub collections. A collection is foundtion of fetching
        and creating new elements.

        :param collection_id: DB id of the collection
        :return:
        """
        return self.get_root_collection().find(id=collection_id)[0]

    def get_collection_of_element(self, element: AbstractElement):
        """
        Find a collection from its element.
        Within this python client it is needed to navigate through all sub collections. A collection is foundtion of fetching
        and creating new elements.

        :param element: A element (Sample, Reaction, ...) aliasable by your account (needs to be saved!)
        :return:
        """
        try:
            try:
                collection_id = element.json_data['tag']['taggable_data']['collection_labels'][0]["id"]
            except (KeyError, IndexError):
                element.load()
                collection_id = element.json_data['tag']['taggable_data']['collection_labels'][0]["id"]
            return self.get_collection_by_id(collection_id)
        except KeyError:
            raise KeyError('Could not find collection of {}'.format(element))

    def search_sample(self) -> SampleSearcher:
        """
        This function returns a SampleSearcher object, which facilitates the search for samples.
        The user has the option to set the maximum number of results and the collection to search in.
        In addition, search conditions can be added and connected with 'or' or 'and'. It is necessary
        to specify the matching type for each condition. Upon making the final call to the method,
        the search will be executed.

        Usage::

        >>> from chemotion_api import Instance
        >>> import logging
        >>> from chemotion_api.search.utils import EnumMatch, EnumConnector
        >>>
        >>> try:
        >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
        >>> except ConnectionError as e:
        >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
        >>> (instance.search_sample().
        >>>  add_search_condition(EnumMatch.LIKE, EnumConnector.OR, name='My Sample', private_note="My note").
        >>>  add_search_condition(EnumMatch.EXACT, EnumConnector.AND, short_label="USER").
        >>>  find_max(5).
        >>>  request())

        :return: SampleSearcher
        """
        return SampleSearcher(self._con, self.element_manager, self.get_root_collection().id)

    def search_reaction(self) -> ReactionSearcher:
        """
        This function returns a ReactionSearcher object, which facilitates the search for reaction.
        The user has the option to set the maximum number of results and the collection to search in.
        In addition, search conditions can be added and connected with 'or' or 'and'. It is necessary
        to specify the matching type for each condition. Upon making the final call to the method,
        the search will be executed.

        Usage::

        >>> from chemotion_api import Instance
        >>> import logging
        >>> from chemotion_api.search.utils import EnumMatch, EnumConnector
        >>>
        >>> try:
        >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
        >>> except ConnectionError as e:
        >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
        >>> (instance.ReactionSearcher().
        >>>  add_search_condition(EnumMatch.LIKE, EnumConnector.OR, name='My Sample', private_note="My note").
        >>>  add_search_condition(EnumMatch.EXACT, EnumConnector.AND, short_label="USER").
        >>>  add_search_condition(EnumMatch.BIGGER_THAN, EnumConnector.AND, temperature_in_c=30).
        >>>  find_max(5).
        >>>  request())

        :return: ReactionSearcher
        """
        return ReactionSearcher(self._con, self.element_manager, self.get_root_collection().id)

    @property
    def all_element_classes(self) -> dict[str, dict[str, str]]:
        """
        This all_element_classes fetches all information about all elements such as
         Sample, Reaction, Wellplate and Research plan and all generic elements

        :return: a dictionary which contains all information about all elements

        :raise RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        return self.element_manager.all_classes

    def get_reaction(self, id: int) -> Reaction:
        """
        Fetches data of one Reaction object from the Chemotion server.
        It automatically parses the data into a Python-Reaction-Object. However, you need to know the correct internally used ID
        of the Reaction to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param id: The Database ID of the desired Element
        :return: a Reacton object
        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        e = ElementSet(self._con, self.all_element_classes.get('reaction'))
        return typing.cast(Reaction, e.load_element(id))

    def get_wellplate(self, id: int) -> Wellplate:
        """
        Fetches data of one Wellplate object from the Chemotion server.
        It automatically parses the data into a Python-Wellplate-Object. However, you need to know the correct internally used ID
        of the Wellplate to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param id: The Database ID of the desired Element
        :return: a Reacton object
        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        e = ElementSet(self._con, self.all_element_classes.get('wellplate'))
        return typing.cast(Wellplate, e.load_element(id))

    def get_research_plan(self, id: int) -> ResearchPlan:
        """
        Fetches data of one Research Plan object from the Chemotion server.
        It automatically parses the data into a Python-ResearchPlan-Object. However, you need to know the correct internally used ID
        of the Research Plan to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param id: The Database ID of the desired Element
        :return: a ResearchPlan object
        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        e = ElementSet(self._con, self.all_element_classes.get('research_plan'))
        return typing.cast(ResearchPlan, e.load_element(id))

    def get_sample(self, id: int) -> Sample:
        """
        Fetches data of one Sample object from the Chemotion server.
        It automatically parses the data into a Python-Sample-Object. However, you need to know the correct internally used ID
        of the Sample to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param id: The Database ID of the desired Element
        :return: a Sample object
        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        e = ElementSet(self._con, self.all_element_classes.get('sample'))
        return typing.cast(Sample, e.load_element(id))

    def get_generic_by_name(self, name: str, id: int) -> GenericElement:
        """
        Fetches data of one Generic object from the Chemotion server. Which generic element type
        It automatically parses the data into a Python-Generic-Object. However, you need to know the correct internally used ID
        of the Generic element to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param name: The name of the Genetic Element
        :param id: The Database ID of the desired Element
        :return: a Sample object

        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        elem = self.all_element_classes.get(name)
        if elem is None:
            raise ValueError(f'Could not find a generic element under the name: "{name}"')
        e = ElementSet(self._con, elem)
        return typing.cast(GenericElement, e.load_element(id))

    def get_generic_by_label(self, label: str, id: int) -> GenericElement:
        """
        Fetches data of one Generic object from the Chemotion server. Which generic element type
        It automatically parses the data into a Python-Generic-Object. However, you need to know the correct internally used ID
        of the Generic element to be able to fetch it. Other methods to get Elements from Chemotion
        are accessible via the Collection objects

        :param label: The label of the Genetic Element
        :param id: The Database ID of the desired Element
        :return: a Sample object

        :raises RequestException: (requests.exceptions.RequestException) if the information cannot be fetched. Make sure that your connection is active and you are logged in.
        """
        for (elem_name, elem) in self.all_element_classes.items():
            if elem['label'] == label:
                return self.get_generic_by_name(elem_name, id)
        raise ValueError(f'Could not find a generic element with the label: "{label}"')

    def get_json_ld_id(self, id: str) -> AbstractElement:
        """
        Use python client internal JSON LD id to fetch element from Chemtion.
        :param id: JSON LD ID string starting with host url of the element
        :return: the requested element

        :raises ValueError: if the information cannot be fetched. Make sure that your connection is active and the ID is correct.
        """
        o = urlparse(id)
        if not self._con.host_url.startswith(f'{o.scheme}://{o.hostname}'):
            raise ValueError('Host dose not fit!')
        path_array = o.path.split('/')
        elem = self.all_element_classes.get(path_array[-2][:-1])
        if elem is None:
            raise ValueError(f'Could not find a generic element under the name: "{path_array[-2]}"')
        e = ElementSet(self._con, elem)
        return e.load_element(int(path_array[-1][:-5]))

    def molecule(self) -> MoleculeManager:
        """
        Returns a MoleculeManager instance. This instance allows to create and
        fetch Molecules form the instance
        :return: a MoleculeManager instance
        """
        return MoleculeManager(self._con)

    def get_solvent_list(self) -> list[str]:
        """
        Returns a list of all pre implemented solvents.
        :return: a list of all pre implemented solvents
        """
        return list(ElementManager.get_solvent_list().keys())

    def generic_manager(self) -> GenericManager:
        """
        Returns a GenericManager used to load, edit write generic elements.
        :return: a GenericManager
        """
        return GenericManager(self._con)



import os.path

import click
import configparser

from chemotion_api.commands import InstanceManager


def deactivate_prompts(ctx, param, value):
    if value:
        for p in ctx.command.params:
            if isinstance(p, click.Option) and p.prompt is not None:
                p.prompt = None
    return value


def read_config(ctx, param, value):
    if isinstance(value, str):
        config = configparser.ConfigParser()
        config.read(value)
        i = config['INSTANCE']
        for p in ctx.command.params:
            for opt in p.opts:
                opt = str(opt).lstrip('-')
                if opt in i:
                    p.default = i[opt]
    return value


def add_options():
    cmd_options = [
        click.option('-q/--quiet', is_flag=True, default=False, is_eager=True, expose_value=False,
                     callback=deactivate_prompts, help='Supress user prompts'),
        click.option('--config', is_eager=True, expose_value=False, callback=read_config, help='Path to config file')
    ]

    def _add_options(func):
        for option in reversed(cmd_options):
            func = option(func)
        return func

    return _add_options


def add_db_options():
    cmd_options = [
        click.option('--pguser', default='postgres', help="Postgresql User"),
        click.option('--port', default='5432', help="Postgresql server port"),
        click.option('--host', default='localhost', help="Postgresql server host"),
        click.option('--pgpassword', default='postgres', help="Postgresql password"),
        click.option('--database', default='chemotion', help="Postgresql database")
    ]

    def _add_options(func):
        for option in reversed(cmd_options):
            func = option(func)
        return func

    return _add_options


@click.group()
def cli():
    pass


@cli.command(help="Fetch all available Version of Chemotion")
@add_options()
def available_versions():
    a = InstanceManager.get_versions()
    click.echo('\n'.join(a))


@cli.command(help="Check all running containers")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@add_options()
def check(docker_compose_path):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    InstanceManager(df).check()


@cli.command(help="Same as docker compose up <services>")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@add_options()
@click.argument('services', nargs=-1, type=click.UNPROCESSED)
def up(docker_compose_path, services):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    services_obj = {'eln': True, 'db': True, 'worker': True, 'ketchersvc': True, 'msconvert': True, 'spectra': True,
                    'converter': True}
    if len(services) != 0:
        for k, v in services_obj.items():
            services_obj[k] = k in services

    InstanceManager(df).up(**services_obj)
    click.echo('Docker-compose is up!')


@cli.command(help="Same as docker compose stop <services>")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@add_options()
@click.argument('services', nargs=-1, type=click.UNPROCESSED)
def stop(docker_compose_path, services):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    services_obj = {'eln': True, 'db': True, 'worker': True, 'ketchersvc': True, 'msconvert': True, 'spectra': True,
                    'converter': True}
    if len(services) != 0:
        for k, v in services_obj.items():
            services_obj[k] = k in services

    InstanceManager(df).stop(**services_obj)
    click.echo('Docker-compose is stoped!')


@cli.command(help="Same as docker compose down <services>")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@add_options()
@click.argument('services', nargs=-1, type=click.UNPROCESSED)
def down(docker_compose_path, services):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    services_obj = {'eln': True, 'db': True, 'worker': True, 'ketchersvc': True, 'msconvert': True, 'spectra': True,
                    'converter': True}
    if len(services) != 0:
        for k, v in services_obj.items():
            services_obj[k] = k in services

    InstanceManager(df).down(**services_obj)
    click.echo('Docker-compose is down!')


@cli.command(help="Create a new docker-compose file with all necessary files and directories for a new chemotion instance.")
@click.option('--ignore_exist', is_flag=True,
              help='If set it will not raise an error in case the target folder is not empty.', default=False)
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@click.option('--version', '-v', prompt='Specify a version',
              help='Chemcommandotion version to be used', default='latest')
@click.option('--url', prompt='Public ELN url',
              help='Public URL on which the ELN can be reached ', default='http://localhost:<ELN_PORT>')
@click.option('--eln_port', '-p', prompt='Port on which the ELN can be reached',
              help='Port on which the ELN can be reached from the host', default=4000)
@click.option('--postgres_port', prompt='Port on which the Postgres DB can be reached',
              help='Port on which the PostgreSQL DB can be reached from the host', default=5432)
@click.option('--adminer_port', prompt='Port on which the Adminer can be reached',
              help='Port on which the Adminer can be reached from the host', default=8080)
@add_options()
def new(ignore_exist, docker_compose_path, version, url, eln_port, postgres_port, adminer_port):
    if url == 'http://localhost:<ELN_PORT>':
        url = str(url).replace('<ELN_PORT>', str(eln_port))
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    try:
        InstanceManager(df).new_instance(version=version, url=url, eln_port=eln_port, postgres_port=postgres_port, adminer_port=adminer_port)
    except ValueError as e:
        if not ignore_exist:
            raise click.ClickException(e.__str__())
        return
    click.echo('Docker-compose file has been created')


@cli.command(help="Change the port on which the ELN can be reached from the host.")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@click.option('--eln_port', '-p', prompt='Port on which the ELN can be reached',
              help='Port on which the ELN can be reached from the host', default=4000)
@add_options()
def change_port(docker_compose_path, eln_port):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    InstanceManager(df).update_port(eln_port=eln_port)
    click.echo('Docker-compose file has been updated (To activate the change, the instance must be restarted)')


@cli.command(help="Change the port on which the ELN DB can be reached from the host.")
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@click.option('--postgres_port', prompt='Port on which the ELN can be reached',
              help='Port on which the Postgresql DB can be reached from the host', default=5432)
@add_options()
def change_postgres_port(docker_compose_path, postgres_port):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    InstanceManager(df).update_db_port(postgresql_port=postgres_port)
    click.echo('Docker-compose file has been updated (To activate the change, the instance must be restarted)')


@cli.command(help='Updated the version of Chemotion instance')
@click.option('--docker_compose_path', '-df', prompt='Your name',
              help='Path to docker compose directory', default='./docker-compose.yml')
@click.option('--version', '-v', prompt='Specify a version',
              help='Chemotion version to be used', default='latest')
@add_options()
def update_version(docker_compose_path, version):
    df = os.path.abspath(os.path.join(os.getcwd(), docker_compose_path))
    instance = InstanceManager(df)
    instance.down()
    instance.update(version=version)
    click.echo('Docker-compose file has been updated')


@cli.command(help="Make a dump of the Chemotion PostgreSQL")
@click.option('--docker_compose_path', '-df',
              help="The path to the Docker Compose directory is not required. "
                   "However, either the specified docker_compose_path or the current "
                   "working directory have to lead to a docker-compose.yml file, the "
                   "script automatically retrieves the PostgreSQL port specified "
                   "in the Docker Compose file, utilizing it as the default value "
                   "for the port argument.", default='./docker-compose.yml')
@click.option('--tar_file', '-f', prompt='Output tar dump file',
              help='Path to output tar dump file', required=False, default='<db_name>_<time_stamp>.tar')
@add_options()
@add_db_options()
def backup(docker_compose_path, tar_file, **kwargs):
    if tar_file == '<db_name>_<time_stamp>.tar':
        tar_file = None
    im = InstanceManager(docker_compose_path)
    if im.check_file():
        im.up(db=True)
        kwargs['port'] = im.db_port
    im.backup(file_path=tar_file, **kwargs)


@cli.command(help="Restore a ProstgreSQL dump. Make sure it is a valid dump in the tar format.")
@click.option('--docker_compose_path', '-df',
              help="The path to the Docker Compose directory is not required. "
                   "However, either the specified docker_compose_path or the current "
                   "working directory have to lead to a docker-compose.yml file, the "
                   "script automatically retrieves the PostgreSQL port specified "
                   "in the Docker Compose file, utilizing it as the default value "
                   "for the port argument.", default='./docker-compose.yml')
@click.option('--tar_file', '-f', prompt='Your tar dump file',
              help='Path to tar dump file')
@add_options()
@add_db_options()
def restore(docker_compose_path, tar_file, **kwargs):
    im = InstanceManager(docker_compose_path)
    if im.check_file():
        im.up(db=True)
        kwargs['port'] = im.db_port
    try:
        im.restore(tar_file, **kwargs)
    except ChildProcessError as e:
        raise click.ClickException(e.__str__())


@cli.command(help="Insert a dataset.")
@click.option('--docker_compose_path', '-df',
              help="The path to the Docker Compose directory is not required. "
                   "However, either the specified docker_compose_path or the current "
                   "working directory have to lead to a docker-compose.yml file, the "
                   "script automatically retrieves the PostgreSQL port specified "
                   "in the Docker Compose file, utilizing it as the default value "
                   "for the port argument.", default='./docker-compose.yml')
@click.option('--ols_term_id', '-ols',  prompt='OLS term id',
              help='OLS (OntoLogy Search) term ID CHMO:XXXX')
@add_options()
@add_db_options()
def insert_dataset(docker_compose_path, ols_term_id, **kwargs):
    im = InstanceManager(docker_compose_path)

    if im.check_file():
        im.up(db=True)
        kwargs['port'] = im.db_port
    try:
        im.insert_dataset(ols_term_id, **kwargs)
    except ChildProcessError as e:
        raise click.ClickException(e.__str__())


if __name__ == '__main__':
    cli()



from chemotion_api.labImotion.items.options import FieldUnit, get_unit
from chemotion_api.labImotion.structure import StructureRegistry

class Select:
    def __init__(self, sr: StructureRegistry, json_data):
        self._option_layers = None
        self.option_layers = json_data.get("option_layers")
        if not hasattr(self, '_sr'):
            self._sr = sr


    @property
    def option_layers(self) -> str:
        return self._option_layers

    @option_layers.setter
    def option_layers(self, value: str | None):
        options = self._sr.get_select_options_list()
        if value is None:
            value = options[0]
        if not isinstance(value, str):
            raise TypeError(f"option_layers {type(value)} not supported")
        if value != '' and value not in options:
            raise ValueError(f"option no in {self._sr.get_select_options_list()}")
        self._option_layers = value

class SystemDefined:
    def __init__(self, json_data: dict):
        self._value_system = None
        self._option_layers = None
        self.option_layers = json_data.get("option_layers", FieldUnit.ACCELERATION)
        self.value_system = json_data.get("value_system", get_unit(self.option_layers)[0])

    @property
    def value_system(self) -> str:
        available_units = get_unit(self.option_layers)
        if self._value_system not in available_units:
            self._value_system = available_units[0]
        return self._value_system

    @value_system.setter
    def value_system(self, value: str):
        available_units = get_unit(self.option_layers)
        if not isinstance(value, str):
            raise TypeError(f"value_system {type(value)} not supported")
        if value not in available_units:
            raise ValueError(f"value_system {value} must be in {available_units}")
        self._value_system = value

    @property
    def option_layers(self) -> FieldUnit:
        if self._option_layers is None:
            return FieldUnit.ACCELERATION
        return self._option_layers

    @option_layers.setter
    def option_layers(self, value: FieldUnit | str):
        if isinstance(value, str):
            value = FieldUnit(value)
        if not isinstance(value, FieldUnit):
            raise TypeError(f"option_layers {type(value)} not supported")
        self._option_layers = value
        available_units = get_unit(self._option_layers)
        if self._value_system not in available_units:
            self._value_system = available_units[0]


import copy
import json
import uuid

from chemotion_api.connection import Connection
from chemotion_api.elements.sample import MoleculeManager
import os.path
from chemotion_api.user import User

from chemotion_api.elements.empty_elements import init_container

from requests.exceptions import RequestException
from chemotion_api.utils.solvent_manager import get_solvent_list


class ElementManager:

    def __init__(self, session: Connection):
        self._session = session
        self._all_classes = None
        self._all_types = None
        self.is_loaded = False

    @property
    def all_classes(self) -> dict[str, dict]:
        if self._all_classes is None:
            self._all_classes = self.get_all_classes()
            self.is_loaded = True
        return self._all_classes


    @property
    def all_types(self) -> list[str]:
        """
        Generates a list af all JSON LD types in the system.
        :return: List of JSON LD types
        """
        if self._all_types is None:
            self._all_types = []
            for name, versions in  self.all_version_of_generic_class.items():
                for v in versions:
                    self._all_types.append(self._session.schema_manager.generate_model_type(name, v))
            self._all_types.sort(reverse=True)

        return self._all_types

    @property
    def all_version_of_generic_class(self) -> dict[str, set[str]]:
        version_list = {}
        for name, class_obj in self.all_classes.items():
            if not class_obj['is_generic']:
                version_list[name] = {None}
            elif 'id' in class_obj:
                version_list[name] = set()
                res = self._session.get(f"/api/v1/generic_elements/klass_revisions.json?id={class_obj['id']}&klass=ElementKlass")
                if res.status_code != 200:
                    raise RequestException('Counld not get the genetic elements')
                for revision in res.json()['revisions']:
                    if revision['version'] is not None:
                        version_list[name].add(revision['version'])
        return version_list


    def get_all_classes(self):
        get_url = "/api/v1/generic_elements/klasses.json"
        res = self._session.get(get_url)
        if res.status_code != 200:
            raise RequestException('Counld not get the genetic elements')
        all_classes = {}
        for x in res.json()['klass']:
            all_classes[x['name']] = x
        all_classes['generic_element'] = {
            'is_generic': True,
            'name': 'generic_element',
            'label': 'generic_element'
        }

        return all_classes

    def _init_container(self):
        return init_container()

    def _get_user(self):
        u = User.load_me(self._session)
        return u

    @staticmethod
    def get_next_shot_label(self, type_name, session):
        return self._get_short_label(type_name)

    def _get_short_label(self, type_name):
        return self._get_user().get_next_short_label(type_name)

    def build_new(self, type_name, collection_id):
        class_obj = self.all_classes[type_name]
        data = {}
        if not class_obj['is_generic']:
            json_path = os.path.join(os.path.dirname(__file__), 'elements/empty_elements', type_name + '.json')
            if os.path.exists(json_path):
                with open(json_path, 'r') as f:
                    data = json.loads(f.read())
            if type_name == 'wellplate':
                data['wells'] = []
                for y in range(1, 9):
                    for x in range(1, 13):
                        data['wells'].append({
                            "id": uuid.uuid4().__str__(),
                            "is_new": True,
                            "type": "well",
                            "position": {
                                "x": x,
                                "y": y
                            },
                            "readouts": [],
                            "sample": None
                        })
        else:
            res = self._session.get('api/v1/generic_elements/klass.json', params={'name': type_name})
            klass_obj = res.json().get('klass')
            properties_release = copy.deepcopy(klass_obj['properties_release'])
            for key, layer in klass_obj['properties_release']['layers'].items():
                if not layer['wf']:
                    properties_release['layers'][key]['ai'] = []
                else:
                    del properties_release['layers'][key]
            data = {
                "type": type_name,
                "is_new": True,
                "name": f"New {klass_obj['label']}",
                "can_copy": True,
                "klassType": "GenericEl",
                "element_klass": klass_obj,
                "element_klass_id": klass_obj['id'],
                "properties": properties_release,
                "properties_release": klass_obj['properties_release'],
                "attachments": [],
                "files": [],
            }

        data['container'] = self._init_container()
        data['collection_id'] = collection_id
        data['short_label'] = self._get_short_label(type_name)
        data['segments'] = []
        return data

    def build_solvent_sample(self, name, collection_id):
        solvent_info = self.get_solvent_list().get(name)
        if solvent_info is None:
            raise KeyError(
                'Solver: "{}" is not available. Run instance.get_solvent_list() to see all valid solver names'.format(
                    name))
        sample_data = self.build_new('sample', collection_id)

        mol = MoleculeManager(self._session).create_molecule_by_smiles(solvent_info['smiles'], solvent_info['density'])

        sample_data['molecule'] = mol
        sample_data['density'] = solvent_info['density']
        sample_data['external_label'] = '{}'.format(name)
        sample_data['short_label'] = 'solvent'
        sample_data['show_label'] = False
        sample_data['reference'] = False
        sample_data['is_split'] = True

        return sample_data

    @classmethod
    def get_solvent_list(cls):
        return get_solvent_list()



from __future__ import annotations  # Enables forward references

import weakref
from _weakref import ReferenceType
from typing import TYPE_CHECKING

from chemotion_api.labImotion.items.options import FieldType

if TYPE_CHECKING:
    from chemotion_api.labImotion.items.genericproperties import GenericProperties
    from chemotion_api.labImotion import GenericObj


class StructureRegistry:
    def __init__(self, obj: GenericObj):
        self._obj: ReferenceType[GenericObj] = weakref.ref(obj)

    def get_properties(self) -> GenericProperties | None:
        return self._obj().properties

    def get_structure(self) -> dict[str, dict[str, FieldType]]:
        res = {}
        props = self.get_properties()
        if props is None:
            for _, layer in self._obj()._json_data['properties_template']['layers'].items():
                res[layer['key']] = {}
                for field in layer['fields']:
                    res[layer['key']][field['field']] = FieldType(field['type'])
        else:
            layers = props.layers
            for l in layers:
                res[l.key] = {}
                for f in l.fields:
                    res[l.key][f.field] = f.field_type
        return res

    def get_select_options_list(self) -> list[str]:
        props = self.get_properties()
        if props is None:
            select_items = self._obj()._json_data['properties_template']['select_options']
        else:
            select_items = props.select_options
        return [key for key, val in select_items.items()]


    def get_select_options(self) -> dict[str: list[str]]:
        props = self.get_properties()
        if props is None:
            select_items = self._obj()._json_data['properties_template']['select_options']
        else:
            select_items = props.select_options
        return {key: {o['key']:o['label'] for o in val['options']} for key, val in select_items.items()}



from chemotion_api.labImotion.items.options import FieldType
from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

generic_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/field/draft-01",
    "title": "Schema for generic field",
    "type": "object",
    "properties": {
        "cols": {
            "oneOf": [
                {
                    "type": "number",
                    "exclusiveMinimum": 1
                },
                {
                    "type": "string",
                    "pattern": "^[2-9]\\d*(\\.\\d+)?$|^1\\.\\d+$"
                }
            ]
        },
        "type": {
            "type": "string",
            "enum": FieldType.list()
        },
        "field": {
            "type": "string"
        },
        "label": {
            "type": "string"
        },
        "default": {
            "type": "string"
        },
        "position": {
            "type": "number",
            "minimum": 0
        },
        "required": {
            "type": "boolean"
        },
        "readonly": {
            "type": "boolean"
        },
        "canAdjust": {
            "type": "boolean"
        },
        "description": {
            "type": "string"
        },
        "sub_fields": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {
                        "type": "string"
                    },
                    "type": {
                        "type": "string"
                    },
                    "value": {
                        "type": "string"
                    },
                    "value_system": {
                        "type": "string"
                    },
                    "option_layers": {
                        "type": "string"
                    },
                    "col_name": {
                        "type": "string"
                    }
                },
                "additionalProperties": False,
                "required": [
                    "id",
                    "type",
                    "value"
                ]
            }
        },
        "text_sub_fields": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {
                        "type": "string"
                    },
                    "field": {
                        "type": "string"
                    },
                    "layer": {
                        "type": "string"
                    },
                    "separator": {
                        "type": "string"
                    }
                },
                "additionalProperties": False,
                "required": [
                    "id",
                    "field",
                    "layer",
                    "separator"
                ]
            }
        },
        "cond_fields": {
            "type": "array",
            "item": {
                "$ref": "chemotion://generic/cond_field/draft-01"
            }
        },
        "cond_operator": {
            "type": "number"
        },
        "decimal": {
            "type": "string"
        },
        "formula": {
            "type": "string"
        },
        "option_layers": {
            "type": "string"
        },
        "value_system": {
            "type": "string"
        },
        "value": {
            "type": ["string", "number", "null"]
        }
    },
    "additionalProperties": False,
    "required": [
        "type",
        "field",
        "label",
        "default",
        "position",
        "sub_fields",
        "text_sub_fields"
    ]

}

SchemaRegistry.instance().register(generic_schema)



import chemotion_api.labImotion.items.validation.schemas.schema_generc_segments
import chemotion_api.labImotion.items.validation.schemas.schema_generc_element
import chemotion_api.labImotion.items.validation.schemas.schema_generc_layer
import chemotion_api.labImotion.items.validation.schemas.schema_generc_linked_element
import chemotion_api.labImotion.items.validation.schemas.schema_select_options
import chemotion_api.labImotion.items.validation.schemas.schema_properties
import chemotion_api.labImotion.items.validation.schemas.schema_cond_field
import chemotion_api.labImotion.items.validation.schemas.schema_flow_obj
import chemotion_api.labImotion.items.validation.schemas.schema_field
import chemotion_api.labImotion.items.validation.schemas.schema_generc_dataset


import copy
from enum import Enum

from chemotion_api.labImotion.items.condition_handler import ConditionHandler
from chemotion_api.labImotion.items.options import FieldType
from chemotion_api.labImotion.structure import StructureRegistry

from chemotion_api.labImotion.utils import all_instances_of, ChemotionSerializable


class BaseField(ChemotionSerializable, ConditionHandler):
    _default_required_fields = ["field_type", "field", "required", "label", "default", "position", "sub_fields",
                                "text_sub_fields"]
    _default_optional_fields = ["cols", "cond_fields", "cond_operator"]
    required_fields = []
    optional_fields = []
    field_type = FieldType.TEXT
    mapping = {"field_type": "type"}

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        ConditionHandler.__init__(self, sr)
        self._field = None
        self._label = None
        self._json_data = json_data
        self._default = ""
        self._position = 1
        self._sr = sr
        self._layer_key = layer_key
        self._cols = None

    def _parse_json(self):
        for key, value in self._json_data.items():
            key = self.mapping.get(key, key)
            if hasattr(self, key):
                setattr(self, key, value)

    @property
    def layer_key(self) -> str:
        return self._layer_key

    def to_dict(self) -> dict:
        self._json_data = {}
        for key in self._default_required_fields + self.required_fields:
            json_key = self.mapping.get(key, key)
            value = getattr(self, key)
            if isinstance(value, Enum):
                value = value.value
            if all_instances_of(value, ChemotionSerializable):
                self._json_data[json_key] = [x.to_dict() for x in value]
            else:
                self._json_data[json_key] = value

        for key in self._default_optional_fields + self.optional_fields:
            json_key = self.mapping.get(key, key)
            value = getattr(self, key, None)
            if value is not None:
                if isinstance(value, Enum):
                    value = value.value
                self._json_data[json_key] = value

        return self._json_data

    @property
    def cols(self) -> int:
        return self._cols

    @cols.setter
    def cols(self, value: int):
        if not isinstance(value, int):
            try:
                value = int(value)
            except ValueError:
                raise TypeError(f"Required {type(value)} not supported")
        self._cols = value

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value: str):
        pass

    @property
    def field(self) -> str | None:
        return self._field

    @field.setter
    def field(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"Required {type(value)} not supported")
        if self.field is None:
            self._field = value
        else:
            raise ValueError("Field is already set and cannot be changed")

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"Required {type(value)} not supported")
        self._label = value

    @property
    def position(self) -> int:
        return self._position

    @position.setter
    def position(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f"Required {type(value)} not supported")
        self._position = value

    @property
    def required(self) -> bool:
        return False

    @required.setter
    def required(self, value: bool):
        pass

    @property
    def sub_fields(self) -> list:
        return []

    @sub_fields.setter
    def sub_fields(self, value: list):
        pass

    @property
    def text_sub_fields(self) -> list[dict]:
        return []

    @text_sub_fields.setter
    def text_sub_fields(self, value: list[dict]):
        pass



import re
import string
from enum import Enum

from chemotion_api.labImotion.items.condition_handler import ConditionHandler
from chemotion_api.labImotion.items.fields import BaseField, field_factory, FieldType, create_field
from chemotion_api.labImotion.items.options import LayerColor
from chemotion_api.labImotion.items.validation import validate_generic_layer
from chemotion_api.labImotion.structure import StructureRegistry
from chemotion_api.labImotion.utils import ChemotionSerializable, all_instances_of


class Layer(ChemotionSerializable, ConditionHandler):
    @classmethod
    def generate_new_layer(cls, sr: StructureRegistry, key: str, label: str):
        return cls(sr, {
            "key": key,
            "label": label,
            "color": "none",
            "style": "panel_generic_heading",
            "cols": 1,
            "position": 10,
            "wf": False,
            "timeRecord": "",
            "wf_position": 0,
            "fields": []
        })

    def __init__(self, sr: StructureRegistry, json_data: dict):
        ConditionHandler.__init__(self, sr)
        self._json_data = json_data
        self.validate()
        self._wf = False
        self._key = ''
        self._style = 'panel_generic_heading'
        self._label = ''
        self._cols = 1
        self._color = ''
        self._position = 1
        self._wf_position = 1
        self._timeRecord = ''
        self._fields = []
        self._sr = sr

        for key, value in json_data.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def add_field(self, field_type: FieldType, label: str, field: str = "", idx: int = -1):
        all_keys = [l.field for l in self.fields]
        if field == '':
            field = re.sub(r'[^a-z]', '_', label.lower())
            t_field = field
            i = 0
            while field in all_keys or field.endswith('_'):
                field = t_field + string.ascii_lowercase[i]
                i += 1

        new_field = create_field(self._sr, self.key, field_type, field, idx, field)
        if idx < 0 :
            self.fields.append(new_field)
        else:
            self.fields.insert(idx, new_field)
        self.update_fields_position()
        return new_field

    def update_fields_position(self):
        i = 1
        for f in self.fields:
            f.position = i
            i += 1


    def validate(self, json_data: dict = None):
        if json_data is None:
            json_data = self._json_data
        validate_generic_layer(json_data)

    def to_dict(self) -> dict:
        for key, _ in self._json_data.items():
            if hasattr(self, key):
                value = getattr(self, key)
                if isinstance(value, Enum):
                    value = value.value
                if all_instances_of(value, ChemotionSerializable):
                    self._json_data[key] = [x.to_dict() for x in value]
                else:
                    self._json_data[key] = value
        return self._json_data

    @property
    def fields(self) -> list[BaseField]:
        return self._fields

    @fields.setter
    def fields(self, value: list[dict | BaseField]):
        self._fields = []
        for field in value:
            if isinstance(field, BaseField):
                self._fields.append(value)
            elif isinstance(field, dict):
                self._fields.append(field_factory(self._sr, self.key, field))

    @property
    def timeRecord(self) -> str:
        return self._timeRecord

    @timeRecord.setter
    def timeRecord(self, time_rec):
        pass

    @property
    def wf(self) -> bool:
        return self._wf

    @wf.setter
    def wf(self, wf: bool):
        if isinstance(wf, bool):
            self._wf = wf
        else:
            raise TypeError('wf must be a boolean')

    @property
    def key(self) -> str:
        return self._key

    @key.setter
    def key(self, key: str):
        if isinstance(key, str):
            self._key = key
        else:
            raise TypeError('key must be a string')

    @property
    def style(self) -> str:
        return self._style

    @style.setter
    def style(self, style: str):
        if re.match(r'^panel_generic_heading$|^panel_generic_heading_([bui])(?!.*\1)([bui])?(?!.*\1|.*\2)([bui])?$',
                    style) is not None:
            self._style = style
        else:
            raise TypeError('style must be a string')

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, label: str):
        if isinstance(label, str):
            self._label = label
        else:
            raise TypeError('label must be a string')

    @property
    def color(self) -> str:
        return self._color

    @color.setter
    def color(self, color: str | LayerColor):
        if isinstance(color, str):
            color = LayerColor(color)
        if isinstance(color, LayerColor):
            self._color = color
        else:
            raise TypeError(f'color must be a string and in [{", ".join(LayerColor.list())}]')

    @property
    def cols(self) -> int:
        return self._cols

    @cols.setter
    def cols(self, cols: int):
        if isinstance(cols, int) and cols > 0:
            self._cols = cols
        else:
            raise TypeError('cols must be a integer and cols > 0')

    @property
    def wf_position(self) -> int:
        return self._wf_position

    @wf_position.setter
    def wf_position(self, wf_position: int):
        if isinstance(wf_position, int) and wf_position >= 0:
            self._wf_position = wf_position
        else:
            raise TypeError('wf_position must be a integer and cols >= 0')

    @property
    def position(self) -> int:
        return self._position

    @position.setter
    def position(self, position: int):
        if isinstance(position, int) and position >= 0:
            self._position = position
        else:
            raise TypeError('position must be a integer and position >= 0')
